"""
Ported bits of netharn. This does not include any of the analytic
output-shape-for methods.

Ignore:
    import liberator
    lib = liberator.Liberator()

    import netharn as nh
    lib.add_dynamic(nh.api.Initializer)

    lib.expand(['netharn'])
    print(lib.current_sourcecode())

    lib.add_dynamic(nh.initializers.KaimingNormal)
    print(lib.current_sourcecode())


    import liberator
    lib = liberator.Liberator()
    lib.add_dynamic(nh.util.number_of_parameters)
    print(lib.current_sourcecode())


    import liberator
    lib = liberator.Liberator()
    import netharn as nh
    lib.add_dynamic(nh.data.collate.padded_collate)
    lib.expand(['netharn'])
    print(lib.current_sourcecode())
"""
import torch
import ubelt as ub
from math import gcd
import torch.nn.functional as F
from torch.utils import data as torch_data
import numpy as np  # NOQA

import collections.abc as container_abcs
from six import string_types as string_classes
from six import integer_types as int_classes
import re

numpy_type_map = {
    'float64': torch.DoubleTensor,
    'float32': torch.FloatTensor,
    'float16': torch.HalfTensor,
    'int64': torch.LongTensor,
    'int32': torch.IntTensor,
    'int16': torch.ShortTensor,
    'int8': torch.CharTensor,
    'uint8': torch.ByteTensor
}


def _update_defaults(config, kw):
    config = dict(config)
    for k, v in kw.items():
        if k not in config:
            config[k] = v
    return config


class Optimizer:
    """
    Old netharn.api.Optimizer class. Ideally this is deprecated.
    """

    @staticmethod
    def coerce(config={}, **kw):
        """
        Accepts keywords:
            optimizer / optim :
                can be sgd, adam, adamw, rmsprop

            learning_rate / lr :
                a float

            weight_decay / decay :
                a float

            momentum:
                a float, only used if the optimizer accepts it

            params:
                This is a SPECIAL keyword that is handled differently.  It is
                interpreted by `netharn.hyper.Hyperparams.make_optimizer`.

                In this simplest case you can pass "params" as a list of torch
                parameter objects or a list of dictionaries containing param
                groups and special group options (just as you would when
                constructing an optimizer from scratch). We don't recommend
                this while using netharn unless you know what you are doing
                (Note, that params will correctly change device if the model is
                mounted).

                In the case where you do not want to group parameters with
                different options, it is best practice to simply not specify
                params.

                In the case where you want to group parameters set params to
                either a List[Dict] or a Dict[str, Dict].

                The items / values of this collection should be a dictionary.
                The keys / values of this dictionary should be the per-group
                optimizer options. Additionally, there should be a key "params"
                (note this is a nested per-group params not to be confused with
                the top-level "params").

                Each per-group "params" should be either (1) a list of
                parameter names (preferred), (2) a string that specifies a
                regular expression (matching layer names will be included in
                this group), or (3) a list of parameter objects.

                For example, the top-level params might look like:

                    params={
                        'head': {'lr': 0.003, 'params': '.*head.*'},
                        'backbone': {'lr': 0.001, 'params': '.*backbone.*'},
                        'preproc': {'lr': 0.0, 'params': [
                            'model.conv1', 'model.norm1', , 'model.relu1']}
                    }

                Note that head and backbone specify membership via regular
                expression whereas preproc explicitly specifies a list of
                parameter names.

        Notes:
            pip install torch-optimizer

        Returns:
            Tuple[type, dict]: a type and arguments to construct it

        References:
            https://datascience.stackexchange.com/questions/26792/difference-between-rmsprop-with-momentum-and-adam-optimizers
            https://github.com/jettify/pytorch-optimizer

        CommandLine:
            xdoctest -m /home/joncrall/code/netharn/netharn/api.py Optimizer.coerce

        Example:
            >>> config = {'optimizer': 'sgd', 'params': [
            >>>     {'lr': 3e-3, 'params': '.*head.*'},
            >>>     {'lr': 1e-3, 'params': '.*backbone.*'},
            >>> ]}
            >>> optim_ = Optimizer.coerce(config)

            >>> # xdoctest: +REQUIRES(module:torch_optimizer)
            >>> config = {'optimizer': 'DiffGrad'}
            >>> optim_ = Optimizer.coerce(config, lr=1e-5)
            >>> print('optim_ = {!r}'.format(optim_))
            >>> assert optim_[1]['lr'] == 1e-5

            >>> config = {'optimizer': 'Yogi'}
            >>> optim_ = Optimizer.coerce(config)
            >>> print('optim_ = {!r}'.format(optim_))

            >>> Optimizer.coerce({'optimizer': 'ASGD'})

        TODO:
            - [ ] https://pytorch.org/blog/stochastic-weight-averaging-in-pytorch/
        """
        config = _update_defaults(config, kw)
        key = config.get('optimizer', config.get('optim', 'sgd')).lower()
        lr = config.get('learning_rate', config.get('lr', 3e-3))
        decay = config.get('weight_decay', config.get('decay', 0))
        momentum = config.get('momentum', 0.9)
        params = config.get('params', None)
        # TODO: allow for "discriminative fine-tuning"
        if key == 'sgd':
            cls = torch.optim.SGD
            kw = {
                'lr': lr,
                'weight_decay': decay,
                'momentum': momentum,
                'nesterov': True,
            }
        elif key == 'adam':
            cls = torch.optim.Adam
            kw = {
                'lr': lr,
                'weight_decay': decay,
                # 'betas': (0.9, 0.999),
                # 'eps': 1e-8,
                # 'amsgrad': False
            }
        elif key == 'adamw':
            from torch.optim import AdamW
            cls = AdamW
            kw = {
                'lr': lr,
                'weight_decay': decay,
                # 'betas': (0.9, 0.999),
                # 'eps': 1e-8,
                # 'amsgrad': False
            }
        elif key == 'rmsprop':
            cls = torch.optim.RMSprop
            kw = {
                'lr': lr,
                'weight_decay': decay,
                'momentum': momentum,
                'alpha': 0.9,
            }
        else:
            _lut = {}

            optim_modules = [
                torch.optim,
            ]

            try:
                # Allow coerce to use torch_optimizer package if available
                import torch_optimizer
            except Exception:
                torch_optimizer = None
            else:
                optim_modules.append(torch_optimizer)
                _lut.update({
                    k.lower(): c.__name__
                    for k, c in torch_optimizer._NAME_OPTIM_MAP.items()})

            _lut.update({
                k.lower(): k for k in dir(torch.optim)
                if not k.startswith('_')})

            key = _lut[key]

            cls = None
            for module in optim_modules:
                cls = getattr(module, key, None)
                if cls is not None:
                    defaultkw = default_kwargs(cls)
                    kw = defaultkw.copy()
                    kw.update(ub.dict_isect(config, kw))
                    # Hacks for common cases, otherwise if learning_rate is
                    # given, but only lr exists in the signature, it will be
                    # incorrectly ignored.
                    if 'lr' in kw:
                        kw['lr'] = lr
                    if 'weight_decay' in kw:
                        kw['weight_decay'] = decay
                    break

        if cls is None:
            raise KeyError(key)

        kw['params'] = params
        optim_ = (cls, kw)
        return optim_


class Initializer(object):
    """
    Base class for initializers
    """

    def __call__(self, model, *args, **kwargs):
        return self.forward(model, *args, **kwargs)

    def forward(self, model):
        """
        Abstract function that does the initailization
        """
        raise NotImplementedError('implement me')

    def history(self):
        """
        Initializer methods have histories which are short for algorithms and
        can be quite long for pretrained models
        """
        return None

    def get_initkw(self):
        """
        Initializer methods have histories which are short for algorithms and
        can be quite long for pretrained models
        """
        initkw = self.__dict__.copy()
        # info = {}
        # info['__name__'] = self.__class__.__name__
        # info['__module__'] = self.__class__.__module__
        # info['__initkw__'] = initkw
        return initkw

    @staticmethod
    def coerce(config={}, **kw):
        """
        Accepts 'init', 'pretrained', 'pretrained_fpath', 'leftover', and
        'noli'.

        Args:
            config (dict | str): coercable configuration dictionary.
                if config is a string it is taken as the value for "init".

        Returns:
            Tuple[Initializer, dict]: initializer_ = initializer_cls, kw

        Examples:
            >>> from geowatch.utils.util_netharn import *  # NOQA
            >>> print(ub.urepr(Initializer.coerce({'init': 'noop'})))
            >>> config = {
            ...     'init': 'pretrained',
            ...     'pretrained_fpath': '/fit/nice/untitled'
            ... }
            >>> print(ub.urepr(Initializer.coerce(config)))
            >>> print(ub.urepr(Initializer.coerce({'init': 'kaiming_normal'})))
        """
        if isinstance(config, str):
            config = {
                'init': config,
            }

        config = _update_defaults(config, kw)

        pretrained_fpath = config.get('pretrained_fpath', config.get('pretrained', None))
        init = config.get('initializer', config.get('init', None))

        # Allow init to specify a pretrained fpath
        if isinstance(init, str) and pretrained_fpath is None:
            from os.path import exists
            pretrained_cand = ub.expandpath(init)
            if exists(pretrained_cand):
                pretrained_fpath = pretrained_cand

        config['init'] = init
        config['pretrained_fpath'] = pretrained_fpath
        config['pretrained'] = pretrained_fpath

        if pretrained_fpath is not None:
            config['init'] = 'pretrained'

        import os
        init_verbose = int(os.environ.get('GEOWATCH_INIT_VERBOSE', os.environ.get('WATCH_INIT_VERBOSE', 4)))

        # ---
        initializer_ = None
        if config['init'].lower() in ['kaiming_normal']:
            initializer_ = (KaimingNormal, {
                # initialization params should depend on your choice of
                # nonlinearity in your model. See the Kaiming Paper for details.
                'param': 1e-2 if config.get('noli', 'relu') == 'leaky_relu' else 0,
            })
        elif config['init'] == 'noop':
            initializer_ = (NoOp, {})
        elif config['init'] == 'pretrained':
            from torch_liberator.initializer import Pretrained
            initializer_ = (Pretrained, {
                'fpath': ub.expandpath(config['pretrained_fpath']),
                'leftover': kw.get('leftover', None),
                'mangle': kw.get('mangle', False),
                'association': kw.get('association', None),
                'verbose': init_verbose,
            })
        elif config['init'] == 'cls':
            # Indicate that the model will initialize itself
            # We have to trust that the user does the right thing here.
            pass
        else:
            raise KeyError('Unknown coercable init: {!r}'.format(config['init']))
        return initializer_


class NoOp(Initializer):
    """
    An initializer that does nothing, which is useful when you have initialized
    the weights yourself.

    Example:
        >>> import copy
        >>> self = NoOp()
        >>> model = ToyNet2d()
        >>> old_state = sum(v.sum() for v in model.state_dict().values())
        >>> self(model)
        >>> new_state = sum(v.sum() for v in model.state_dict().values())
        >>> assert old_state == new_state
        >>> assert self.history() is None
    """

    def forward(self, model):
        return


class Orthogonal(Initializer):
    """
    Same as Orthogonal, but uses pytorch implementation

    Example:
        >>> self = Orthogonal()
        >>> model = ToyNet2d()
        >>> try:
        >>>     self(model)
        >>> except RuntimeError:
        >>>     import pytest
        >>>     pytest.skip('geqrf: Lapack probably not availble')
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """

    def __init__(self, gain=1):
        self.gain = gain

    def forward(self, model):
        try:
            func = torch.nn.init.orthogonal_
        except AttributeError:
            func = torch.nn.init.orthogonal

        apply_initializer(model, func, self.__dict__)


class KaimingUniform(Initializer):
    """
    Same as HeUniform, but uses pytorch implementation

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> self = KaimingUniform()
        >>> model = ToyNet2d()
        >>> self(model)
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """

    def __init__(self, param=0, mode='fan_in'):
        self.a = param
        self.mode = mode

    def forward(self, model):
        try:
            func = torch.nn.init.kaiming_uniform_
        except AttributeError:
            func = torch.nn.init.kaiming_uniform
        apply_initializer(model, func, self.__dict__)


class KaimingNormal(Initializer):
    """
    Same as HeNormal, but uses pytorch implementation

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> self = KaimingNormal()
        >>> model = ToyNet2d()
        >>> self(model)
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """

    def __init__(self, param=0, mode='fan_in'):
        self.a = param
        self.mode = mode

    def forward(self, model):
        try:
            func = torch.nn.init.kaiming_normal_
        except AttributeError:
            func = torch.nn.init.kaiming_normal
        apply_initializer(model, func, self.__dict__)


def apply_initializer(input, func, funckw):
    """
    Recursively initializes the input using a torch.nn.init function.

    If the input is a model, then only known layer types are initialized.

    Args:
        input (Tensor | Module): can be a model, layer, or tensor
        func (callable): initialization function
        funckw (dict):

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> from torch import nn
        >>> import torch
        >>> class DummyNet(nn.Module):
        >>>     def __init__(self, n_channels=1, n_classes=10):
        >>>         super(DummyNet, self).__init__()
        >>>         self.conv = nn.Conv2d(n_channels, 10, kernel_size=5)
        >>>         self.norm = nn.BatchNorm2d(10)
        >>>         self.param = torch.nn.Parameter(torch.rand(3))
        >>> self = DummyNet()
        >>> func = nn.init.kaiming_normal_
        >>> apply_initializer(self, func, {})
        >>> func = nn.init.constant_
        >>> apply_initializer(self, func, {'val': 42})
        >>> assert np.all(self.conv.weight.detach().numpy() == 42)
        >>> assert np.all(self.conv.bias.detach().numpy() == 0), 'bias is always init to zero'
        >>> assert np.all(self.norm.bias.detach().numpy() == 0), 'bias is always init to zero'
        >>> assert np.all(self.norm.weight.detach().numpy() == 1)
        >>> assert np.all(self.norm.running_mean.detach().numpy() == 0.0)
        >>> assert np.all(self.norm.running_var.detach().numpy() == 1.0)
    """
    if getattr(input, 'bias', None) is not None:
        # print('zero input bias')
        # zero all biases
        input.bias.data.zero_()

    if isinstance(input, (torch.Tensor)):
        # assert False, ('input is tensor? does this make sense?')
        # print('input is tensor')
        func(input, **funckw)
        # data = input
    elif isinstance(input, (torch.nn.modules.conv._ConvNd)):
        # print('input is convnd')
        func(input.weight, **funckw)
    # elif isinstance(input, (torch.nn.modules.linear.Linear)):
    #     func(input.weight, **funckw)
    elif isinstance(input, torch.nn.modules.batchnorm._BatchNorm):
        # Use default batch norm
        input.reset_parameters()
    # elif isinstance(input, torch.nn.modules.Linear):
    #     input.reset_parameters()
    elif hasattr(input, 'reset_parameters'):
        # print('unknown input type fallback on reset_params')
        input.reset_parameters()
    else:
        # input is a torch module
        model = input
        # print('recurse input')
        layers = list(trainable_layers(model))
        # print('layers = {!r}'.format(layers))
        for item in layers:
            apply_initializer(item, func, funckw)


def trainable_layers(model, names=False):
    """
    Returns all layers containing trainable parameters

    Notes:
        It may be better to simply use model.named_parameters() instead in most
        situation. This is useful when you need the classes that contains the
        parameters instead of the parameters themselves.

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> import torchvision
        >>> model = torchvision.models.AlexNet()
        >>> list(trainable_layers(model, names=True))
    """
    if names:
        stack = [('', '', model)]
        while stack:
            prefix, basename, item = stack.pop()
            name = '.'.join([p for p in [prefix, basename] if p])
            if isinstance(item, torch.nn.modules.conv._ConvNd):
                yield name, item
            elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
                yield name, item
            elif hasattr(item, 'reset_parameters'):
                yield name, item

            child_prefix = name
            for child_basename, child_item in list(item.named_children())[::-1]:
                stack.append((child_prefix, child_basename, child_item))
    else:
        queue = [model]
        while queue:
            item = queue.pop(0)
            # TODO: need to put all trainable layer types here
            # (I think this is just everything with reset_parameters)
            if isinstance(item, torch.nn.modules.conv._ConvNd):
                yield item
            elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
                yield item
            elif hasattr(item, 'reset_parameters'):
                yield item
            # if isinstance(input, torch.nn.modules.Linear):
            #     yield item
            # if isinstance(input, torch.nn.modules.Bilinear):
            #     yield item
            # if isinstance(input, torch.nn.modules.Embedding):
            #     yield item
            # if isinstance(input, torch.nn.modules.EmbeddingBag):
            #     yield item
            for child in item.children():
                queue.append(child)


def number_of_parameters(model, trainable=True):
    """
    Returns number of trainable parameters in a torch module

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> model = torch.nn.Conv1d(2, 3, 5)
        >>> print(number_of_parameters(model))
        33
    """
    if trainable:
        model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    else:
        model_parameters = model.parameters()
    n_params = sum([np.prod(p.size()) for p in model_parameters])
    return n_params


class ToyNet2d(torch.nn.Module):
    """
    Demo model for a simple 2 class learning problem
    """
    def __init__(self, input_channels=1, num_classes=2):
        super().__init__()
        self.input_channels = input_channels
        self.num_classes = num_classes
        self.layers = torch.nn.Sequential(*[
            torch.nn.Conv2d(input_channels, 8, kernel_size=3, padding=1, bias=False),

            torch.nn.BatchNorm2d(8),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(8, 8, kernel_size=3, padding=1, bias=False),

            torch.nn.BatchNorm2d(8),
            torch.nn.ReLU(inplace=True),
            torch.nn.Conv2d(8, num_classes, kernel_size=3, padding=1, bias=False),
        ])

        self.softmax = torch.nn.Softmax(dim=1)

    def forward(self, inputs):
        spatial_out = self.layers(inputs)
        num = float(np.prod(spatial_out.shape[-2:]))
        averaged = spatial_out.sum(dim=2).sum(dim=2) / num
        probs = self.softmax(averaged)
        return probs


class ToyData2d(torch_data.Dataset):
    """
    Simple black-on-white and white-on-black images.

    Args:
        n (int, default=100): dataset size
        size (int, default=4): width / height
        border (int, default=1): border mode
        rng (RandomCoercable, default=None): seed or random state

    CommandLine:
        python -m netharn.data.toydata ToyData2d --show

    Example:
        >>> self = ToyData2d()
        >>> data1, label1 = self[0]
        >>> data2, label2 = self[-1]
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> kwplot.figure(fnum=1, doclf=True)
        >>> kwplot.imshow(data1.numpy().squeeze(), pnum=(1, 2, 1))
        >>> kwplot.imshow(data2.numpy().squeeze(), pnum=(1, 2, 2))
        >>> kwplot.show_if_requested()
    """
    def __init__(self, size=4, border=1, n=100, rng=None):
        import kwarray
        import itertools as it
        rng = kwarray.ensure_rng(rng)

        h = w = size

        whiteish = 1 - (np.abs(rng.randn(n, 1, h, w) / 4) % 1)
        blackish = (np.abs(rng.randn(n, 1, h, w) / 4) % 1)

        fw = border
        slices = [slice(None, fw), slice(-fw, None)]

        # class 0 is white block inside a black frame
        data1 = whiteish.copy()
        for sl1, sl2 in it.product(slices, slices):
            data1[..., sl1, :] = blackish[..., sl1, :]
            data1[..., :, sl2] = blackish[..., :, sl2]

        # class 1 is black block inside a white frame
        data2 = blackish.copy()
        for sl1, sl2 in it.product(slices, slices):
            data2[..., sl1, :] = whiteish[..., sl1, :]
            data2[..., :, sl2] = whiteish[..., :, sl2]

        self.data = np.concatenate([data1, data2], axis=0)
        self.labels = np.array(([0] * n) + ([1] * n))

        suffix = ub.hash_data([
            size, border, n, rng
        ], base='abc', hasher='sha1')[0:16]
        self.input_id = 'TD2D_{}_'.format(n) + suffix

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        data = torch.FloatTensor(self.data[index])
        label = int(self.labels[index])
        return data, label

    def make_loader(self, *args, **kwargs):
        loader = torch_data.DataLoader(self, *args, **kwargs)
        return loader


class InputNorm(torch.nn.Module):
    """
    Normalizes the input by shifting and dividing by a scale factor.

    This allows for the network to take care of 0-mean 1-std normalization.
    The developer explicitly specifies what these shift and scale values are.
    By specifying this as a layer (instead of a data preprocessing step), the
    exporter will remember and associated this information with any
    deployed model. This means that a user does not need to remember what these
    shit/scale arguments were before passing inputs to a network.

    If the mean and std arguments are unspecified, this layer becomes a noop.

    References:
        .. [WhyNormalize] https://towardsdatascience.com/why-data-should-be-normalized-before-training-a-neural-network-c626b7f66c7d

    Example:
        >>> self = InputNorm(mean=50.0, std=29.0)
        >>> inputs = torch.rand(2, 3, 5, 7) * 100
        >>> outputs = self(inputs)
        >>> # If mean and std are unspecified, this becomes a noop.
        >>> assert torch.all(InputNorm()(inputs) == inputs)
        >>> # Specifying either the mean or the std is ok.
        >>> partial1 = InputNorm(mean=50)(inputs)
        >>> partial2 = InputNorm(std=29)(inputs)

    Ignore:
        import torch

        model = torch.nn.Sequential(*[
            InputNorm(mean=10, std=0.2),
            torch.nn.Conv2d(3, 3, 3),
        ])
        inputs = torch.rand(2, 3, 5, 7) * 100
        optim = torch.optim.SGD(model.parameters(), lr=1e-3)

        for i in range(100):
            optim.zero_grad()
            x = model(inputs).sum()
            x.backward()
            optim.step()

            std = model[0].mean
            mean = model[0].std
            print('std = {!r}'.format(std))
            print('mean = {!r}'.format(mean))
    """

    def __init__(self, mean=None, std=None):
        super(InputNorm, self).__init__()
        if mean is not None:
            mean = mean if ub.iterable(mean) else [mean]
            mean = torch.FloatTensor(mean)
        if std is not None:
            std = std if ub.iterable(std) else [std]
            std = torch.FloatTensor(std)
        # Buffers are not returned by .parameters, so they are not modified by
        # optimization, but they are stored in the state_dict
        self.register_buffer('mean', mean)
        self.register_buffer('std', std)

    def forward(self, inputs):
        outputs = inputs
        if self.mean is not None:
            outputs = outputs - self.mean
        if self.std is not None:
            outputs = outputs / self.std
        return outputs


class MultiLayerPerceptronNd(torch.nn.Module):
    """
    A multi-layer perceptron network for n dimensional data

    Choose the number and size of the hidden layers, number of output channels,
    wheather to user residual connections or not, nonlinearity, normalization,
    dropout, and more.

    Args:
        dim (int): specify if the data is 0, 1, 2, 3, or 4 dimensional.

        in_channels (int): number of input channels

        hidden_channels (List[int]): or an int specifying the number of hidden
            layers (we choose the channel size to linearly interpolate between
            input and output channels)

        out_channels (int): number of output channels

        dropout (float, default=None): amount of dropout to use between 0 and 1

        norm (str, default='batch'): type of normalization layer
            (e.g. batch or group), set to None for no normalization.

        noli (str, default='relu'): type of nonlinearity

        residual (bool, default=False):
            if true includes a resitual skip connection between inputs and
            outputs.

        norm_output (bool, default=True):
            if True, applies a final normalization layer to the output.

        noli_output (bool, default=True):
            if True, applies a final nonlineary to the output.

        standardize_weights (bool, default=False):
            Use weight standardization

    Example:
        >>> kw = {'dim': 0, 'in_channels': 2, 'out_channels': 1}
        >>> model0 = MultiLayerPerceptronNd(hidden_channels=0, **kw)
        >>> model1 = MultiLayerPerceptronNd(hidden_channels=1, **kw)
        >>> model2 = MultiLayerPerceptronNd(hidden_channels=2, **kw)
        >>> print('model0 = {!r}'.format(model0))
        >>> print('model1 = {!r}'.format(model1))
        >>> print('model2 = {!r}'.format(model2))

        >>> kw = {'dim': 0, 'in_channels': 2, 'out_channels': 1, 'residual': True}
        >>> model0 = MultiLayerPerceptronNd(hidden_channels=0, **kw)
        >>> model1 = MultiLayerPerceptronNd(hidden_channels=1, **kw)
        >>> model2 = MultiLayerPerceptronNd(hidden_channels=2, **kw)
        >>> print('model0 = {!r}'.format(model0))
        >>> print('model1 = {!r}'.format(model1))
        >>> print('model2 = {!r}'.format(model2))

    Example:
        >>> import ubelt as ub
        >>> self = MultiLayerPerceptronNd(dim=1, in_channels=128, hidden_channels=3, out_channels=2)
        >>> print(self)
        MultiLayerPerceptronNd...
    """
    def __init__(self, dim, in_channels, hidden_channels, out_channels,
                 bias=True, dropout=None, noli='relu', norm='batch',
                 residual=False, noli_output=False, norm_output=False,
                 standardize_weights=False):

        super(MultiLayerPerceptronNd, self).__init__()
        dropout_cls = rectify_dropout(dim)
        conv_cls = rectify_conv(dim=dim)
        curr_in = in_channels

        if isinstance(hidden_channels, int):
            n = hidden_channels
            hidden_channels = np.linspace(in_channels, out_channels, n + 1,
                                          endpoint=False)[1:]
            hidden_channels = hidden_channels.round().astype(int).tolist()
        self._hidden_channels = hidden_channels

        hidden = self.hidden = torch.nn.Sequential()
        for i, curr_out in enumerate(hidden_channels):
            layer = ConvNormNd(
                dim, curr_in, curr_out, kernel_size=1, bias=bias, noli=noli,
                norm=norm, standardize_weights=standardize_weights)
            hidden.add_module('hidden{}'.format(i), layer)
            if dropout is not None:
                hidden.add_module('dropout{}'.format(i), dropout_cls(p=dropout))
            curr_in = curr_out

        outkw = {'bias': bias, 'kernel_size': 1}
        self.hidden.add_module(
            'output', conv_cls(curr_in, out_channels, **outkw))

        if residual:
            if in_channels == out_channels:
                self.skip = Identity()
            else:
                self.skip = conv_cls(in_channels, out_channels, **outkw)
        else:
            self.skip = None

        if norm_output:
            self.final_norm = rectify_normalizer(out_channels, norm, dim=dim)
        else:
            self.final_norm = None

        if noli_output:
            self.final_noli = rectify_nonlinearity(noli, dim=dim)
        else:
            self.final_noli = None

        self.norm_output = norm_output
        self.noli_output = noli_output
        self.in_channels = in_channels
        self.out_channels = out_channels

    def forward(self, inputs):
        outputs = self.hidden(inputs)

        if self.skip is not None:
            projected = self.skip(inputs)
            outputs = projected + outputs

        if self.final_norm is not None:
            outputs = self.final_norm(outputs)

        if self.final_noli is not None:
            outputs = self.final_noli(outputs)

        return outputs


def rectify_dropout(dim=2):
    conv_cls = {
        0: torch.nn.Dropout,
        1: torch.nn.Dropout,
        2: torch.nn.Dropout2d,
        3: torch.nn.Dropout3d,
    }[dim]
    return conv_cls


def rectify_nonlinearity(key=ub.NoParam, dim=2):
    """
    Allows dictionary based specification of a nonlinearity

    Example:
        >>> rectify_nonlinearity('relu')
        ReLU(...)
        >>> rectify_nonlinearity('leaky_relu')
        LeakyReLU(negative_slope=0.01...)
        >>> rectify_nonlinearity(None)
        None
        >>> rectify_nonlinearity('swish')
    """
    if key is None:
        return None

    if key is ub.NoParam:
        key = 'relu'

    if isinstance(key, str):
        key = {'type': key}
    elif isinstance(key, dict):
        key = key.copy()
    else:
        raise TypeError(type(key))
    kw = key
    noli_type = kw.pop('type')
    if 'inplace' not in kw:
        kw['inplace'] = True

    if noli_type == 'leaky_relu':
        cls = torch.nn.LeakyReLU
    elif noli_type == 'relu':
        cls = torch.nn.ReLU
    elif noli_type == 'elu':
        cls = torch.nn.ELU
    elif noli_type == 'celu':
        cls = torch.nn.CELU
    elif noli_type == 'selu':
        cls = torch.nn.SELU
    elif noli_type == 'relu6':
        cls = torch.nn.ReLU6
    elif noli_type == 'swish':
        kw.pop('inplace', None)
        cls = Swish
    elif noli_type == 'mish':
        kw.pop('inplace', None)
        cls = Mish
    else:
        raise KeyError('unknown type: {}'.format(kw))
    return cls(**kw)


def rectify_normalizer(in_channels, key=ub.NoParam, dim=2, **kwargs):
    """
    Allows dictionary based specification of a normalizing layer

    Args:
        in_channels (int): number of input channels
        dim (int): dimensionality
        **kwargs: extra args

    Example:
        >>> rectify_normalizer(8)
        BatchNorm2d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, 'batch')
        BatchNorm2d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, {'type': 'batch'})
        BatchNorm2d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, 'group')
        GroupNorm(4, 8, eps=1e-05, affine=True)
        >>> rectify_normalizer(8, {'type': 'group', 'num_groups': 2})
        GroupNorm(2, 8, eps=1e-05, affine=True)
        >>> rectify_normalizer(8, dim=3)
        BatchNorm3d(8, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        >>> rectify_normalizer(8, None)
        None
        >>> rectify_normalizer(8, key={'type': 'syncbatch'})
        >>> rectify_normalizer(8, {'type': 'group', 'num_groups': 'auto'})
        >>> rectify_normalizer(1, {'type': 'group', 'num_groups': 'auto'})
        >>> rectify_normalizer(16, {'type': 'group', 'num_groups': 'auto'})
        >>> rectify_normalizer(32, {'type': 'group', 'num_groups': 'auto'})
        >>> rectify_normalizer(64, {'type': 'group', 'num_groups': 'auto'})
        >>> rectify_normalizer(1024, {'type': 'group', 'num_groups': 'auto'})
    """
    if key is None:
        return None

    if key is ub.NoParam:
        key = 'batch'

    if isinstance(key, str):
        key = {'type': key}
    elif isinstance(key, dict):
        key = key.copy()
    else:
        raise TypeError(type(key))

    norm_type = key.pop('type')
    if norm_type == 'batch':
        in_channels_key = 'num_features'

        if dim == 0:
            cls = torch.nn.BatchNorm1d
        elif dim == 1:
            cls = torch.nn.BatchNorm1d
        elif dim == 2:
            cls = torch.nn.BatchNorm2d
        elif dim == 3:
            cls = torch.nn.BatchNorm3d
        else:
            raise ValueError(dim)
    elif norm_type == 'syncbatch':
        in_channels_key = 'num_features'
        cls = torch.nn.SyncBatchNorm
    elif norm_type == 'group':
        in_channels_key = 'num_channels'
        if key.get('num_groups') is None:
            key['num_groups'] = 'auto'
            # key['num_groups'] = ('gcd', min(in_channels, 32))

        if key.get('num_groups') == 'auto':
            if in_channels == 1:
                # Warning: cant group norm this
                return Identity()
            else:
                valid_num_groups = [
                    factor for factor in range(1, in_channels)
                    if in_channels % factor == 0
                ]
                if len(valid_num_groups) == 0:
                    raise Exception
                infos = [
                    {'ng': ng, 'nc': in_channels / ng}
                    for ng in valid_num_groups
                ]
                ideal = in_channels ** (0.5)
                for item in infos:
                    item['heuristic'] = abs(ideal - item['ng']) * abs(ideal - item['nc'])
                chosen = sorted(infos, key=lambda x: (x['heuristic'], 1 - x['ng']))[0]
                key['num_groups'] = chosen['ng']
                if key['num_groups'] == in_channels:
                    key['num_groups'] = 1

            if isinstance(key['num_groups'], tuple):
                if key['num_groups'][0] == 'gcd':
                    key['num_groups'] = gcd(
                        key['num_groups'][1], in_channels)

            if in_channels % key['num_groups'] != 0:
                raise AssertionError(
                    'Cannot divide n_inputs {} by num groups {}'.format(
                        in_channels, key['num_groups']))
        cls = torch.nn.GroupNorm

    elif norm_type == 'batch+group':
        return torch.nn.Sequential(
            rectify_normalizer(in_channels, 'batch', dim=dim),
            rectify_normalizer(in_channels, ub.dict_union({'type': 'group'}, key), dim=dim),
        )
    else:
        raise KeyError('unknown type: {}'.format(key))
    assert in_channels_key not in key
    key[in_channels_key] = in_channels

    try:
        import copy
        kw = copy.copy(key)
        kw.update(kwargs)
        return cls(**kw)
    except Exception:
        raise
        # Ignore kwargs
        import warnings
        warnings.warn('kwargs ignored in rectify normalizer')
        return cls(**key)


def _ws_extra_repr(self):
    s = ('{in_channels}, {out_channels}, kernel_size={kernel_size}'
         ', stride={stride}')
    if self.padding != (0,) * len(self.padding):
        s += ', padding={padding}'
    if self.dilation != (1,) * len(self.dilation):
        s += ', dilation={dilation}'
    if self.output_padding != (0,) * len(self.output_padding):
        s += ', output_padding={output_padding}'
    if self.groups != 1:
        s += ', groups={groups}'
    if self.bias is None:
        s += ', bias=False'
    if self.padding_mode != 'zeros':
        s += ', padding_mode={padding_mode}'
    if self.standardize_weights:
        s += ', standardize_weights={standardize_weights}'
    return s.format(**self.__dict__)


class Identity(torch.nn.Sequential):
    """
    A identity-function layer.

    Example:
        >>> import torch
        >>> self = Identity()
        >>> a = torch.rand(3, 3)
        >>> b = self(a)
        >>> assert torch.all(a == b)
    """
    def __init__(self):
        super(Identity, self).__init__()


class Conv0d(torch.nn.Linear):
    """
    Ignore:
        self = Conv0d(2, 3, 1, standardize_weights=True)
        print('self = {!r}'.format(self))
        x = torch.rand(1, 2)
        y = self.forward(x)
        print('y = {!r}'.format(y))
    """
    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1,
                 padding=0, dilation=1, groups=1, bias=True,
                 padding_mode='zeros', standardize_weights=False):
        assert kernel_size == 1, 'Conv0D must have a kernel_size=1'
        assert padding == 0, 'Conv0D must have padding=1'
        assert stride == 1, 'Conv0D must have stride=1'
        assert groups == 1, 'Conv0D must have groups=1'
        assert dilation == 1, 'Conv0D must have a dilation=1'
        # assert padding_mode == 'zeros'
        super().__init__(in_features=in_channels, out_features=out_channels,
                         bias=bias)
        self.standardize_weights = standardize_weights
        if standardize_weights:
            assert in_channels > 1, 'must be greater than 1 to prevent nan'
        self.dim = 0
        self.eps = 1e-5

    def forward(self, x):
        if self.standardize_weights:
            weight = weight_standardization_nd(self.dim, self.weight, self.eps)
            return torch.nn.functional.linear(x, weight, self.bias)
        else:
            return super().forward(x)

    def extra_repr(self) -> str:
        s = 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )
        if self.standardize_weights:
            s += ', standardize_weights={standardize_weights}'
        return s


class Conv1d(torch.nn.Conv1d):
    """
    Ignore:
        self = Conv1d(2, 3, 1, standardize_weights=True)
        print('self = {!r}'.format(self))
        x = torch.rand(1, 2, 1)
        y = self.forward(x)
        print('y = {!r}'.format(y))
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True,
                 standardize_weights=False):
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            groups, bias)
        self.standardize_weights = standardize_weights
        if standardize_weights:
            assert in_channels > 1, 'must be greater than 1 to prevent nan'
        self.eps = 1e-5
        self.dim = 1

    def forward(self, x):
        if self.standardize_weights:
            weight = weight_standardization_nd(self.dim, self.weight, self.eps)
            return torch.nn.functional.conv1d(
                x, weight, self.bias, self.stride, self.padding, self.dilation,
                self.groups)
        else:
            return super().forward(x)

    extra_repr = _ws_extra_repr


class Conv2d(torch.nn.Conv2d):
    """
    Ignore:
        self = Conv2d(2, 3, 1, standardize_weights=True)
        print('self = {!r}'.format(self))
        x = torch.rand(1, 2, 3, 3)
        y = self.forward(x)
        print('y = {!r}'.format(y))
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True,
                 standardize_weights=False):
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            groups, bias)
        self.standardize_weights = standardize_weights
        if standardize_weights:
            assert in_channels > 1, 'must be greater than 1 to prevent nan'
        self.eps = 1e-5
        self.dim = 2

    def forward(self, x):
        if self.standardize_weights:
            weight = weight_standardization_nd(self.dim, self.weight, self.eps)
            return torch.nn.functional.conv2d(
                x, weight, self.bias, self.stride, self.padding, self.dilation,
                self.groups)
        else:
            return super().forward(x)

    extra_repr = _ws_extra_repr


class Conv3d(torch.nn.Conv3d):
    """
    Ignore:
        self = Conv3d(2, 3, 1, standardize_weights=True)
        print('self = {!r}'.format(self))
        x = torch.rand(1, 2, 1, 1, 1)
        y = self.forward(x)
        print('y = {!r}'.format(y))
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True,
                 standardize_weights=False):
        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            groups, bias)
        self.standardize_weights = standardize_weights
        if standardize_weights:
            assert in_channels > 1, 'must be greater than 1 to prevent nan'
        self.eps = 1e-5
        self.dim = 3

    def forward(self, x):
        if self.standardize_weights:
            weight = weight_standardization_nd(self.dim, self.weight, self.eps)
            return torch.nn.functional.conv3d(
                x, weight, self.bias, self.stride, self.padding, self.dilation,
                self.groups)
        else:
            return super().forward(x)

    extra_repr = _ws_extra_repr


def rectify_conv(dim=2):
    conv_cls = {
        0: Conv0d,
        # 1: torch.nn.Conv1d,
        # 2: torch.nn.Conv2d,
        # 3: torch.nn.Conv3d,
        1: Conv1d,
        2: Conv2d,
        3: Conv3d,
    }[dim]
    return conv_cls


def weight_standardization_nd(dim, weight, eps):
    """
    Note: input channels must be greater than 1!

    Ignore:
        weight = torch.rand(3, 2, 1, 1)
        dim = 2
        eps = 1e-5
        weight_normed = weight_standardization_nd(dim, weight, eps)
        print('weight = {!r}'.format(weight))
        print('weight_normed = {!r}'.format(weight_normed))

        weight = torch.rand(1, 2)
        dim = 0
        eps = 1e-5
        weight_normed = weight_standardization_nd(dim, weight, eps)
        print('weight = {!r}'.format(weight))
        print('weight_normed = {!r}'.format(weight_normed))
    """
    # Note: In 2D Weight dimensions are [C_out, C_in, H, W]
    mean_dims = tuple(list(range(1, dim + 2)))
    weight_mean = weight.mean(dim=mean_dims, keepdim=True)
    weight = weight - weight_mean
    trailing = [1] * (dim + 1)
    std = weight.view(weight.shape[0], -1).std(dim=1).view(-1, *trailing) + eps
    weight = weight / std.expand_as(weight)
    return weight


class ConvNormNd(torch.nn.Sequential):
    """
    Backbone convolution component. The convolution hapens first, normalization
    and nonlinearity happen after the convolution.

    CONV[->NORM][->NOLI]

    Args:
        dim (int):
            dimensionality of the convolutional kernel (can be 0, 1, 2, or 3).

        in_channels (int):

        out_channels (int):

        kernel_size (int | Tuple):

        stride (int | Tuple):

        padding (int | Tuple):

        dilation (int | Tuple):

        groups (int):

        bias (bool):

        norm (str, dict, nn.Module): Type of normalizer,
            if None, then normalization is disabled.

        noli (str, dict, nn.Module): Type of nonlinearity,
            if None, then normalization is disabled.

        standardize_weights (bool, default=False):
            Implements weight standardization as described in Qiao 2020 -
            "Micro-Batch Training with Batch-Channel Normalization and Weight
            Standardization"- https://arxiv.org/pdf/1903.10520.pdf

    Example:
        >>> self = ConvNormNd(dim=2, in_channels=16, out_channels=64,
        >>>                    kernel_size=3)
        >>> print(self)
        ConvNormNd(
          (conv): Conv2d(16, 64, kernel_size=(3, 3), stride=(1, 1))
          (norm): BatchNorm2d(64, ...)
          (noli): ReLU(...)
        )

    Example:
        >>> self = ConvNormNd(dim=0, in_channels=16, out_channels=64)
        >>> print(self)
        ConvNormNd(
          (conv): Conv0d(in_features=16, out_features=64, bias=True)
          (norm): BatchNorm1d(64, ...)
          (noli): ReLU(...)
        )
        >>> input_shape = (None, 16)
    """
    def __init__(self, dim, in_channels, out_channels, kernel_size=1, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, noli='relu',
                 norm='batch', standardize_weights=False):
        super(ConvNormNd, self).__init__()

        conv_cls = rectify_conv(dim)
        conv = conv_cls(in_channels, out_channels, kernel_size=kernel_size,
                        padding=padding, stride=stride, groups=groups,
                        bias=bias, dilation=dilation,
                        standardize_weights=standardize_weights)

        norm = rectify_normalizer(out_channels, norm, dim=dim)
        noli = rectify_nonlinearity(noli, dim=dim)

        self.add_module('conv', conv)
        if norm:
            self.add_module('norm', norm)
        if noli:
            self.add_module('noli', noli)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.standardize_weights = standardize_weights
        self._dim = dim


class ConvNorm1d(ConvNormNd):
    """
    Backbone convolution component. The convolution hapens first, normalization
    and nonlinearity happen after the convolution.

    CONV[->NORM][->NOLI]

    Args:
        norm (str, dict, nn.Module): Type of normalizer,
            if None, then normalization is disabled.
        noli (str, dict, nn.Module): Type of nonlinearity,
            if None, then normalization is disabled.

    Example:
        >>> input_shape = [2, 3, 5]
        >>> self = ConvNorm1d(input_shape[1], 7, kernel_size=3)
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, noli='relu',
                 norm='batch', standardize_weights=False):
        super(ConvNorm1d, self).__init__(dim=1, in_channels=in_channels,
                                         out_channels=out_channels,
                                         kernel_size=kernel_size,
                                         stride=stride, bias=bias,
                                         padding=padding, noli=noli, norm=norm,
                                         dilation=dilation, groups=groups,
                                         standardize_weights=standardize_weights)


class ConvNorm2d(ConvNormNd):
    """
    Backbone convolution component. The convolution hapens first, normalization
    and nonlinearity happen after the convolution.

    CONV[->NORM][->NOLI]

    Args:
        norm (str, dict, nn.Module): Type of normalizer,
            if None, then normalization is disabled.
        noli (str, dict, nn.Module): Type of nonlinearity,
            if None, then normalization is disabled.

    Example:
        >>> input_shape = [2, 3, 5, 7]
        >>> self = ConvNorm2d(input_shape[1], 11, kernel_size=3)
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, noli='relu',
                 norm='batch', standardize_weights=False):
        super(ConvNorm2d, self).__init__(dim=2, in_channels=in_channels,
                                         out_channels=out_channels,
                                         kernel_size=kernel_size,
                                         stride=stride, bias=bias,
                                         padding=padding, noli=noli, norm=norm,
                                         dilation=dilation, groups=groups,
                                         standardize_weights=standardize_weights)


class ConvNorm3d(ConvNormNd):
    """
    Backbone convolution component. The convolution hapens first, normalization
    and nonlinearity happen after the convolution.

    CONV[->NORM][->NOLI]

    Args:
        norm (str, dict, nn.Module): Type of normalizer,
            if None, then normalization is disabled.
        noli (str, dict, nn.Module): Type of nonlinearity,
            if None, then normalization is disabled.

    Example:
        >>> input_shape = [2, 3, 5, 7, 11]
        >>> self = ConvNorm3d(input_shape[1], 13, kernel_size=3)
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 bias=True, padding=0, noli='relu', norm='batch',
                 groups=1, standardize_weights=False):
        super(ConvNorm3d, self).__init__(dim=3, in_channels=in_channels,
                                         out_channels=out_channels,
                                         kernel_size=kernel_size,
                                         stride=stride, bias=bias,
                                         padding=padding, noli=noli, norm=norm,
                                         groups=groups,
                                         standardize_weights=standardize_weights)


class _SwishFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, i):
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        i = ctx.saved_variables[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))


class Swish(torch.nn.Module):
    """
    When beta=1 this is Sigmoid-weighted Linear Unit (SiL)

    ``x * torch.sigmoid(x)``

    References:
        https://arxiv.org/pdf/1710.05941.pdf

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> x = torch.linspace(-20, 20, 100, requires_grad=True)
        >>> self = Swish()
        >>> y = self(x)
        >>> y.sum().backward()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.multi_plot(xydata={'beta=1': (x.data, y.data)}, fnum=1, pnum=(1, 2, 1),
        >>>         ylabel='swish(x)', xlabel='x', title='activation')
        >>> kwplot.multi_plot(xydata={'beta=1': (x.data, x.grad)}, fnum=1, pnum=(1, 2, 2),
        >>>         ylabel='𝛿swish(x) / 𝛿(x)', xlabel='x', title='gradient')
        >>> kwplot.show_if_requested()

    """
    def __init__(self, beta=1.0):
        super(Swish, self).__init__()
        self.beta = beta

    def forward(self, x):
        """
        Equivalent to ``x * torch.sigmoid(x)``
        """
        if self.beta == 1:
            return _SwishFunction.apply(x)
        else:
            return x * torch.sigmoid(x * self.beta)


@torch.jit.script
def mish(input):
    """
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    See additional documentation for mish class.
    """
    return input * torch.tanh(F.softplus(input))


def beta_mish(input, beta=1.5):
    """
    Applies the β mish function element-wise:
        .. math::
            \\beta mish(x) = x * tanh(ln((1 + e^{x})^{\\beta}))
    See additional documentation for :mod:`echoAI.Activation.Torch.beta_mish`.

    References:
        https://github.com/digantamisra98/Echo/blob/master/echoAI/Activation/Torch/functional.py
    """
    return input * torch.tanh(torch.log(torch.pow((1 + torch.exp(input)), beta)))


class Mish_Function(torch.autograd.Function):
    """
    Applies the mish function element-wise:

    Math:
        mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + e^{x}))

    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input

    References:
        https://github.com/digantamisra98/Echo/blob/master/echoAI/Activation/Torch/mish.py

    Examples:
        >>> m = Mish()
        >>> input = torch.randn(2)
        >>> output = m(input)
    """

    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        y = x * torch.tanh(F.softplus(x))  # x * tanh(ln(1 + exp(x)))
        return y

    @staticmethod
    def backward(ctx, grad_output):
        x = ctx.saved_variables[0]
        sigmoid = torch.sigmoid(x)
        tanh_sp = torch.tanh(F.softplus(x))
        return grad_output * (tanh_sp + x * sigmoid * (1 - tanh_sp * tanh_sp))

    # else:
    #     @torch.jit.script
    #     def mish(input):
    #         delta = torch.exp(-input)
    #         alpha = 1 + 2 * delta
    #         return input * alpha / (alpha + 2 * delta * delta)


class Mish(torch.nn.Module):
    """
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))

    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input

    References:
        https://github.com/digantamisra98/Mish/blob/master/Mish/Torch/mish.py
        https://github.com/thomasbrandon/mish-cuda
        https://arxiv.org/pdf/1908.08681v2.pdf

    Examples:
        >>> m = Mish()
        >>> input = torch.randn(2)
        >>> output = m(input)

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> x = torch.linspace(-20, 20, 100, requires_grad=True)
        >>> self = Mish()
        >>> y = self(x)
        >>> y.sum().backward()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.multi_plot(xydata={'beta=1': (x.data, y.data)}, fnum=1, pnum=(1, 2, 1))
        >>> kwplot.multi_plot(xydata={'beta=1': (x.data, x.grad)}, fnum=1, pnum=(1, 2, 2))
        >>> kwplot.show_if_requested()
    """
    def __init__(self):
        super().__init__()

    def forward(self, input):
        return Mish_Function.apply(input)
        # return mish(input)


def _debug_inbatch_shapes(inbatch):
    import ubelt as ub
    print('len(inbatch) = {}'.format(len(inbatch)))
    extensions = ub.util_format.FormatterExtensions()
    #
    @extensions.register((torch.Tensor, np.ndarray))
    def format_shape(data, **kwargs):
        return ub.repr2(dict(type=str(type(data)), shape=data.shape), nl=1, sv=1)
    print('inbatch = ' + ub.repr2(inbatch, extensions=extensions, nl=True))


def default_kwargs(cls):
    """
    Grab initkw defaults from the constructor

    Args:
        cls (type | callable): a class or function

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> import torch
        >>> import ubelt as ub
        >>> cls = torch.optim.Adam
        >>> default_kwargs(cls)
        >>> cls = KaimingNormal
        >>> print(ub.repr2(default_kwargs(cls), nl=0))
        {'mode': 'fan_in', 'param': 0}
        >>> cls = NoOp
        >>> default_kwargs(cls)
        {}

    SeeAlso:
        xinspect.get_func_kwargs(cls)
    """
    import inspect
    sig = inspect.signature(cls)
    default_kwargs = {
        k: p.default
        for k, p in sig.parameters.items()
        if p.default is not p.empty
    }
    return default_kwargs


default_collate = torch_data.dataloader.default_collate


def _collate_else(batch, collate_func):
    """
    Handles recursion in the else case for these special collate functions

    This is duplicates all non-tensor cases from `torch_data.dataloader.default_collate`
    This also contains support for collating slices.
    """
    error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
    elem_type = type(batch[0])
    if elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
            and elem_type.__name__ != 'string_':
        elem = batch[0]
        if elem_type.__name__ == 'ndarray':
            # array of string classes and object
            if re.search('[SaUO]', elem.dtype.str) is not None:
                raise TypeError(error_msg.format(elem.dtype))

            return torch.stack([torch.from_numpy(b) for b in batch], 0)
        if elem.shape == ():  # scalars
            py_type = float if elem.dtype.name.startswith('float') else int
            return numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
    elif isinstance(batch[0], slice):
        batch = default_collate([{
            'start': sl.start,
            'stop': sl.stop,
            'step': 1 if sl.step is None else sl.step
        } for sl in batch])
        return batch
    elif isinstance(batch[0], int_classes):
        return torch.LongTensor(batch)
    elif isinstance(batch[0], float):
        return torch.DoubleTensor(batch)
    elif isinstance(batch[0], string_classes):
        return batch
    elif isinstance(batch[0], container_abcs.Mapping):
        if 0:
            # Hack the mapping collation implementation to print error info
            collated = {}
            try:
                for key in batch[0]:
                    collated[key] = collate_func([d[key] for d in batch])
            except Exception:
                print('\n!!Error collating key = {!r}\n'.format(key))
                raise
            return collated
        else:
            return {key: collate_func([d[key] for d in batch]) for key in batch[0]}
    elif isinstance(batch[0], tuple) and hasattr(batch[0], '_fields'):  # namedtuple
        return type(batch[0])(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(batch[0], container_abcs.Sequence):
        transposed = zip(*batch)
        return [collate_func(samples) for samples in transposed]
    else:
        raise TypeError((error_msg.format(type(batch[0]))))


class CollateException(Exception):
    pass


def padded_collate(inbatch, fill_value=-1):
    """
    Used for detection datasets with boxes.

    Example:
        >>> from geowatch.utils.util_netharn import *  # NOQA
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 7
        >>> for i in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     n = 11 if i == 3 else rng.randint(0, 11)
        >>>     boxes = torch.rand(n, 4)
        >>>     item = (img, boxes)
        >>>     inbatch.append(item)
        >>> out_batch = padded_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> assert list(out_batch[1].shape) == [bsize, 11, 4]

    Example:
        >>> import torch
        >>> rng = np.random.RandomState(0)
        >>> inbatch = []
        >>> bsize = 4
        >>> for _ in range(bsize):
        >>>     # add an image and some dummy bboxes to the batch
        >>>     img = torch.rand(3, 8, 8)  # dummy 8x8 image
        >>>     #boxes = torch.empty(0, 4)
        >>>     boxes = torch.FloatTensor()
        >>>     item = (img, [boxes])
        >>>     inbatch.append(item)
        >>> out_batch = padded_collate(inbatch)
        >>> assert len(out_batch) == 2
        >>> assert list(out_batch[0].shape) == [bsize, 3, 8, 8]
        >>> #assert list(out_batch[1][0].shape) == [bsize, 0, 4]
        >>> assert list(out_batch[1][0].shape) in [[0], []]  # torch .3 a .4

    Example:
        >>> inbatch = [torch.rand(4, 4), torch.rand(8, 4),
        >>>            torch.rand(0, 4), torch.rand(3, 4),
        >>>            torch.rand(0, 4), torch.rand(1, 4)]
        >>> out_batch = padded_collate(inbatch)
        >>> assert list(out_batch.shape) == [6, 8, 4]
    """
    try:
        if torch.is_tensor(inbatch[0]):
            num_items = [len(item) for item in inbatch]
            if ub.allsame(num_items):
                if len(num_items) == 0:
                    batch = torch.FloatTensor()
                elif num_items[0] == 0:
                    batch = torch.FloatTensor()
                else:
                    batch = default_collate(inbatch)
            else:
                max_size = max(num_items)
                real_tail_shape = None
                for item in inbatch:
                    if item.numel():
                        tail_shape = item.shape[1:]
                        if real_tail_shape is not None:
                            assert real_tail_shape == tail_shape
                        real_tail_shape = tail_shape

                padded_inbatch = []
                for item in inbatch:
                    n_extra = max_size - len(item)
                    if n_extra > 0:
                        shape = (n_extra,) + tuple(real_tail_shape)
                        if torch.__version__.startswith('0.3'):
                            extra = torch.Tensor(np.full(shape, fill_value=fill_value))
                        else:
                            extra = torch.full(shape, fill_value=fill_value,
                                               dtype=item.dtype)
                        padded_item = torch.cat([item, extra], dim=0)
                        padded_inbatch.append(padded_item)
                    else:
                        padded_inbatch.append(item)
                batch = inbatch
                batch = default_collate(padded_inbatch)
        else:
            batch = _collate_else(inbatch, padded_collate)
    except Exception as ex:
        if not isinstance(ex, CollateException):
            try:
                _debug_inbatch_shapes(inbatch)
            except Exception:
                pass
            raise CollateException(
                'Failed to collate inbatch={}. Reason: {!r}'.format(inbatch, ex))
        else:
            raise
    return batch



from performer_pytorch import FastAttention
import math
import einops
from torch import nn
import ubelt as ub

try:
    import xdev
    profile = xdev.profile
except Exception:
    profile = ub.identity


class FastMultiheadSelfAttention(FastAttention):
    """
    This seems like a good idea, but either I'm using it wrong or the
    C-bindings in normal attention make this lose all of its benefit.

    Ignore:
        D = 9  # embedding dimension
        H = 3   # number of heads
        B = 5   # batch size
        S = 7   # sequence length
        x = torch.rand(S, B, D)
        MultiheadSelfAttention(D, H)(x).shape
        FastMultiheadSelfAttention(D, H)(x)
        from performer_pytorch import FastAttention
        q = einops.rearrange(x, 's b (h e) -> b h s e', h=H)
        FastAttention(dim_heads=D // H, nb_features=None)(q, q, q).shape
    """

    def __init__(self, embed_dim, num_heads):
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        assert embed_dim % num_heads == 0
        dim_heads = embed_dim // num_heads
        nb_features = int(dim_heads * math.log(dim_heads))
        # nb_features = int(dim_heads * 2)
        super().__init__(
            dim_heads, nb_features=nb_features, ortho_scaling=0,
            causal=False, generalized_attention=False, kernel_fn=nn.ReLU(),
            no_projection=False)

    @profile
    def forward(self, x, key_padding_mask=None):
        # import xdev
        # xdev.embed()
        # make compatible with nn.MultiheadAttention
        # s, b, he = x.shape
        # e = self.dim_heads
        # h = self.num_heads
        # Much faster than einops
        if key_padding_mask is not None:
            raise NotImplementedError
        # q = x.contiguous().view(s, b, h, e).permute(1, 2, 0, 3)
        q = einops.rearrange(x, 's b (h e) -> b h s e', e=self.dim_heads)
        # a = FastAttention.forward(self, q, q, q)
        a = super().forward(q, q, q)
        # out = a.permute(2, 1, 0, 3).contiguous().view(s, b, he)
        out = einops.rearrange(a, 'b h s e -> s b (h e)', e=self.dim_heads)
        return out



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains utility functions for the model code."""

from typing import Any, List, MutableMapping, MutableSequence, Optional, Set

import tensorflow as tf
from deeplab2 import common
from deeplab2 import config_pb2

layers = tf.keras.layers

_PREDICTION_WITH_NEAREST_UPSAMPLING = (
    common.PRED_INSTANCE_KEY,
    common.PRED_INSTANCE_CENTER_KEY,
    common.PRED_INSTANCE_SCORES_KEY,
    common.PRED_SEMANTIC_SCORES_KEY,
    common.PRED_PANOPTIC_KEY,
    common.PRED_SEMANTIC_KEY,
    common.PRED_CENTER_HEATMAP_KEY,
    common.PRED_NEXT_PANOPTIC_KEY,
    common.PRED_CONCAT_NEXT_PANOPTIC_KEY,
)

_PREDICTION_WITH_BILINEAR_UPSAMPLING = (
    common.PRED_SEMANTIC_PROBS_KEY,
    common.PRED_DEPTH_KEY,
    common.PRED_OFFSET_MAP_KEY,
)

_INPUT_WITH_NEAREST_UPSAMPLING = (
    common.GT_INSTANCE_CENTER_KEY,
)

_INPUT_WITH_BILINEAR_UPSAMPLING = (
    common.IMAGE,
    common.GT_INSTANCE_REGRESSION_KEY
)


def resolve_batch_size(tensor: tf.Tensor) -> int:
    batch_size = tensor.get_shape().as_list()[0]
    if batch_size is None:
        batch_size = tf.shape(tensor)[0]
    return batch_size


def _scale_helper(value, scale):
    if isinstance(value, tf.Tensor):
        return tf.cast(
            (tf.cast(value, dtype=tf.float32) - 1.0) * scale + 1.0,
            dtype=tf.int32)
    else:
        return int((float(value) - 1.0) * scale + 1.0)


def scale_mutable_sequence(input_sequence: MutableSequence[int],
                           scale: float) -> MutableSequence[int]:
    return [_scale_helper(x, scale) for x in input_sequence]


def scale_int_list(int_list, scale):
    return [int(x * scale) for x in int_list]


def undo_image_preprocessing(image_in: tf.Tensor, method: str,
                             perform_crop: bool,
                             regions_to_crop: List[int],
                             output_shape: List[int]) -> tf.Tensor:
    """Undoes the image preprocessing.

    In particular, this function slices out the valid regions (determined by
    `regions_to_crop`) in the input when perform_crop is True. After
    that, we resize the results to the desired `output_shape`.

    Args:
      image_in: Input image Tensor with shape [batch, height, width, n_channels].
      method: Image resize method.
      perform_crop: Boolean, performing crop or not.
      regions_to_crop: The regions to crop [height, width]. Will only apply
        cropping at the bottom right.
      output_shape: Desired shape after resizing [height, width].

    Returns:
      Outputs after cropping (if perform_crop = True) and resizing.
    """
    if perform_crop:
        image_out = image_in[
                    :, :regions_to_crop[0], :regions_to_crop[1], :]
    else:
        image_out = image_in
    return resize_align_corners(image_out, output_shape, method=method)


def undo_preprocessing(input_or_prediction_dict: MutableMapping[str, Any],
                       regions_to_crop: List[int],
                       output_shape: List[int]) -> MutableMapping[str, Any]:
    """Undoes preprocessing for predictions.

    Args:
      input_or_prediction_dict: A dictionary storing different types of inputs or
        predictions.
      regions_to_crop: The regions to crop [height, width]. Will only apply
        cropping at the bottom right.
      output_shape: Desired shape after resizing [height, width].

    Returns:
      inputs or predictions after cropping (if perform_crop = True) and resizing.
    """
    for key in input_or_prediction_dict.keys():
        if key in _PREDICTION_WITH_NEAREST_UPSAMPLING or key in _INPUT_WITH_NEAREST_UPSAMPLING:
            input_or_prediction_dict[key] = tf.squeeze(
                undo_image_preprocessing(
                    tf.expand_dims(input_or_prediction_dict[key], 3),
                    'nearest',
                    perform_crop=True,
                    regions_to_crop=regions_to_crop,
                    output_shape=output_shape),
                axis=3)
        elif key in _PREDICTION_WITH_BILINEAR_UPSAMPLING or key in _INPUT_WITH_BILINEAR_UPSAMPLING:
            input_or_prediction_dict[key] = undo_image_preprocessing(
                input_or_prediction_dict[key],
                'bilinear',
                perform_crop=True,
                regions_to_crop=regions_to_crop,
                output_shape=output_shape)
        else:
            # We only undo preprocessing for those defined in
            # _{PREDICTION,INPUT}_WITH_{NEAREST,BILINEAR}_UPSAMPLING.
            # Other intermediate results are skipped.
            continue
    return input_or_prediction_dict


def add_zero_padding(input_tensor: tf.Tensor, kernel_size: int,
                     rank: int) -> tf.Tensor:
    """Adds zero-padding to the input_tensor."""
    pad_total = kernel_size - 1
    pad_begin = pad_total // 2
    pad_end = pad_total - pad_begin
    if rank == 3:
        return tf.pad(
            input_tensor,
            paddings=[[pad_begin, pad_end], [pad_begin, pad_end], [0, 0]])
    else:
        return tf.pad(
            input_tensor,
            paddings=[[0, 0], [pad_begin, pad_end], [pad_begin, pad_end], [0, 0]])


def resize_and_rescale_offsets(input_tensor: tf.Tensor, target_size):
    """Bilinearly resizes and rescales the offsets.

    Args:
      input_tensor: A tf.Tensor of shape [batch, height, width, 2].
      target_size: A list or tuple or 1D tf.Tensor that specifies the height and
        width after resizing.

    Returns:
      The input_tensor resized to shape `[batch, target_height, target_width, 2]`.
        Moreover, the offsets along the y-axis are rescaled by a factor equal to
        (target_height - 1) / (reference_height - 1) and the offsets along the
        x-axis are rescaled by a factor equal to
        (target_width - 1) / (reference_width - 1).
    """
    input_size_y = tf.shape(input_tensor)[1]
    input_size_x = tf.shape(input_tensor)[2]

    scale_y = tf.cast(target_size[0] - 1, tf.float32) / tf.cast(
        input_size_y - 1, tf.float32)
    scale_x = tf.cast(target_size[1] - 1, tf.float32) / tf.cast(
        input_size_x - 1, tf.float32)

    target_y, target_x = tf.split(
        value=input_tensor, num_or_size_splits=2, axis=3)
    target_y *= scale_y
    target_x *= scale_x
    target = tf.concat([target_y, target_x], 3)
    return resize_bilinear(target, target_size)


def resize_align_corners(input_tensor, target_size, method='bilinear'):
    """Resizes the input_tensor to target_size.

    This returns the same output as tf.compat.v1.image.resize(input_tensor,
    target_size, align_corners=True).

    Args:
      input_tensor: A tf.Tensor of shape [batch, height, width, channels].
      target_size: A list or tuple or 1D tf.Tensor that specifies the height and
        width after resizing.
      method: An optional string specifying the method used for resizing.
        Supported options are 'nearest' and 'bilinear'.

    Returns:
      The resized tensor.

    Raises:
      ValueError: An error occurs if 1) the input tensor's rank is not 4 or 2) the
        resizing method is not supported.
    """
    if method == 'bilinear':
        tf_method = tf.compat.v1.image.ResizeMethod.BILINEAR
    elif method == 'nearest':
        tf_method = tf.compat.v1.image.ResizeMethod.NEAREST_NEIGHBOR
    else:
        raise ValueError('The given method %s is not supported. Please use bilinear'
                         ' or nearest.' % method)

    tf.debugging.assert_rank(
        input_tensor, 4,
        message='Input tensor to resize method should have rank of 4.')

    return tf.compat.v1.image.resize(
        input_tensor,
        target_size,
        method=tf_method,
        align_corners=True,
        name='resize_align_corners')


def resize_bilinear(images,
                    size,
                    align_corners=True,
                    name=None):
    """TPU memory efficient version of tf.compat.v1.image.resize_bilinear.

    ResizeBilinear on TPU requires padded batch and channel dimensions. On a
    TPUv3, the worst case could lead to 256x memory consumption, if the
    input is, for example, [1, 257, 513, 1]. In this function, we replace the
    default resize_bilinear by two resize_bilinear operations, which put one image
    axis on the channel axis. This reduces TPU padding when batch * channel is
    small and height * width is large.

    Args:
      images: Input image of shape [B, H, W, C].
      size: A list of two elements: [height, width]. The new size for the images.
      align_corners: Whether to align corners of the image.
      name: Name of the operation.

    Returns:
      Resized image.
    """
    _, height, width, channel = images.get_shape().as_list()
    if height == size[0] and width == size[1]:
        return images
    dtype = images.dtype
    images = tf.cast(images, tf.float32)
    # We check the channel axis only since the batch size is similar (usually 1 or
    # 2). In this way, this if-else easily supports dynamic batch size without
    # using tf.cond().
    if channel > 32 or not align_corners:
        images = tf.compat.v1.image.resize_bilinear(
            images, size,
            align_corners=align_corners,
            name=name)
    else:
        images = tf.transpose(images, [0, 3, 1, 2])
        images = tf.compat.v1.image.resize_bilinear(
            images, [channel, size[0]],
            align_corners=align_corners,
            name=name + '_height' if name else None)
        images = tf.transpose(images, [0, 1, 3, 2])
        images = tf.compat.v1.image.resize_bilinear(
            images, [channel, size[1]],
            align_corners=align_corners,
            name=name + '_width' if name else None)
        images = tf.transpose(images, [0, 3, 2, 1])
    return tf.cast(images, dtype)


def make_divisible(value: float,
                   divisor: int,
                   min_value: Optional[float] = None) -> int:
    """Ensures all layers have channels that are divisible by the divisor.

    Args:
      value: A `float` of original value.
      divisor: An `int` of the divisor that needs to be checked upon.
      min_value: A `float` of  minimum value threshold.

    Returns:
      The adjusted value in `int` that is divisible by divisor.

    Raises:
      ValueError: Minimual value should be divisible by divisor.
    """
    if min_value is None:
        min_value = divisor
    elif min_value % divisor != 0:
        raise ValueError('Minimual value should be divisible by divisor.')

    new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_value < 0.9 * value:
        new_value += divisor
    return int(new_value)


def transpose_and_reshape_for_attention_operation(inputs):
    """Sequentially transposes and reshapes the tensor.

    Args:
      inputs: An input [batch, num_heads, length, channel] tensor.

    Returns:
      output: An output [batch, length, num_heads * channel] tensor.
    """
    _, num_heads, length, channel = inputs.get_shape().as_list()
    transposed_inputs = tf.transpose(inputs, [0, 2, 1, 3])
    return tf.reshape(transposed_inputs, [-1, length, num_heads * channel])


def reshape_and_transpose_for_attention_operation(inputs, num_heads):
    """Sequentially reshapes and transposes the tensor.

    Args:
      inputs: An input [batch, length, num_heads * channel] tensor.
      num_heads: An integer, the number of attention heads.

    Returns:
      output: An output [batch, num_heads, length, channel] tensor.
    """
    _, length, channels = inputs.get_shape().as_list()
    inputs = tf.reshape(inputs, [-1, length, num_heads, channels // num_heads])
    return tf.transpose(inputs, [0, 2, 1, 3])


def get_layer_name(private_attribute_name):
    if private_attribute_name[0] != '_':
        raise ValueError('Private attribute name should start with a \'_\'.')
    return private_attribute_name[1:]


def get_stem_current_name(index):
    return '_basic_block{}'.format(index + 1)


def get_low_level_conv_fusion_conv_current_names(index):
    return ('_low_level_conv{}'.format(index + 1),
            '_fusion_conv{}'.format(index + 1))


def get_conv_bn_act_current_name(index, use_bn, activation):
    name = '_conv{}'.format(index + 1)
    if use_bn:
        name += '_bn'
    if (activation is not None and
            activation.lower() != 'none' and
            activation.lower() != 'linear'):
        name += '_act'
    return name


def safe_setattr(obj, name, value):
    """A conflict-safe version of setattr().

    Different from setattr(), this function raises ValueError if the object
    already has an attribute with the same name.

    Args:
      obj: An object whose attribute has to be set.
      name: A string, the name of the attribute.
      value: Any type, the value given to the attribute.

    Raises:
      ValueError: If the object already has an attribute with the same name.
    """
    if hasattr(obj, name):
        raise ValueError('The object already has an attribute with the same name.')
    setattr(obj, name, value)


def pad_sequence_with_none(sequence, target_length):
    return list(sequence) + [None] * (target_length - len(sequence))


def strided_downsample(input_tensor, target_size):
    """Strided downsamples a tensor to the target size.

    The stride_height and stride_width is computed by (height - 1) //
    (target_height - 1) and (width - 1) // (target_width - 1). We raise an error
    if stride_height != stride_width, since this is not intended in our current
    use cases. But this check can be removed if different strides are desired.
    This function supports static shape only.

    Args:
      input_tensor: A [batch, height, width] tf.Tensor to be downsampled.
      target_size: A list of two integers, [target_height, target_width], the
        target size after downsampling.

    Returns:
      output_tensor: A [batch, target_height, target_width] tf.Tensor, the
        downsampled result.

    Raises:
      ValueError: If the input cannot be downsampled with integer stride, i.e.,
        (height - 1) % (target_height - 1) != 0, or (width - 1) % (target_width -
        1) != 0.
      ValueError: If the height axis stride does not equal to the width axis
        stride.
    """
    input_height, input_width = input_tensor.get_shape().as_list()[1:3]
    target_height, target_width = target_size

    if ((input_height - 1) % (target_height - 1) or
            (input_width - 1) % (target_width - 1)):
        raise ValueError('The input cannot be downsampled with integer striding. '
                         'Please ensure (height - 1) % (target_height - 1) == 0 '
                         'and (width - 1) % (target_width - 1) == 0.')
    stride_height = (input_height - 1) // (target_height - 1)
    stride_width = (input_width - 1) // (target_width - 1)
    if stride_height != stride_width:
        raise ValueError('The height axis stride does not equal to the width axis '
                         'stride.')
    if stride_height > 1 or stride_width > 1:
        return input_tensor[:, ::stride_height, ::stride_width]
    return input_tensor


def get_num_thing_stuff_classes(num_classes: int, void_label: int) -> int:
    """Computes num_thing_stuff_classes.

    The num_thing_stuff_classes are computed from the num_classes and the
    void_label.

    Args:
      num_classes: An integer specifying the number of semantic classes, which
        may or may not include `void` class.
      void_label: An integer specifying the void label.

    Returns:
      num_thing_stuff_classes: An integer specifying the number of stuff and thing
        classes, excluding `void` class.
    """
    if void_label < 0 or void_label > num_classes:
        num_thing_stuff_classes = num_classes
    else:
        num_thing_stuff_classes = num_classes - 1
    return num_thing_stuff_classes


def get_stuff_class_ids(num_thing_stuff_classes: int,
                        thing_class_ids: List[int],
                        void_label: int) -> List[int]:
    """Computes stuff_class_ids.

    The stuff_class_ids are computed from the num_thing_stuff_classes, the
    thing_class_ids and the void_label.

    Args:
      num_thing_stuff_classes: An integer specifying the number of stuff and thing
        classes, not including `void` class.
      thing_class_ids: A List of integers of length [num_thing_classes] containing
        thing class indices.
      void_label: An integer specifying the void label.

    Returns:
      stuff_class_ids: A sorted List of integers of shape [num_stuff_classes]
        containing stuff class indices.
    """
    if void_label >= num_thing_stuff_classes:
        thing_stuff_class_ids = list(range(num_thing_stuff_classes))
    else:
        thing_stuff_class_ids = [_ for _ in range(num_thing_stuff_classes + 1)
                                 if _ is not void_label]
    return sorted(set(thing_stuff_class_ids) - set(thing_class_ids))


def get_supported_tasks(
        config: config_pb2.ExperimentOptions) -> Set[str]:
    """Gets currently supported tasks for each meta_architecture.

    Args:
      config: A config_pb2.ExperimentOptions configuration.

    Returns:
      supported_tasks: A set of strings (see common.py), optionally
       - common.TASK_PANOPTIC_SEGMENTATION,
       - common.TASK_INSTANCE_SEGMENTATION,
       - common.TASK_VIDEO_PANOPTIC_SEGMENTATION,
    """
    supported_tasks = set()
    meta_architecture = config.model_options.WhichOneof('meta_architecture')
    is_max_deeplab = meta_architecture == 'max_deeplab'
    is_motion_deeplab = meta_architecture == 'motion_deeplab'
    is_panoptic_deeplab = meta_architecture == 'panoptic_deeplab'
    is_vip_deeplab = meta_architecture == 'vip_deeplab'
    is_panoptic = (
            (config.model_options.panoptic_deeplab.instance.enable and
             is_panoptic_deeplab) or
            is_motion_deeplab or is_max_deeplab or is_vip_deeplab)
    if is_panoptic:
        supported_tasks.add(common.TASK_PANOPTIC_SEGMENTATION)
        supported_tasks.add(common.TASK_INSTANCE_SEGMENTATION)
    if is_motion_deeplab or is_vip_deeplab:
        supported_tasks.add(common.TASK_VIDEO_PANOPTIC_SEGMENTATION)
    if is_vip_deeplab and config.model_options.vip_deeplab.HasField('depth_head'):
        supported_tasks.add(common.TASK_DEPTH_AWARE_VIDEO_PANOPTIC_SEGMENTATION)
    return supported_tasks



#!/usr/bin/env python3
"""
SeeAlso:
    ~/code/watch-smartflow-dags/KIT_TA2_PREEVAL10_PYENV.py
"""
import os
import shutil
import sys
import traceback
import scriptconfig as scfg
import ubelt as ub



__debugging__ = r"""

LOCAL_WORK_DPATH=$HOME/temp/debug_smartflow_v2_kr2-sv/ingress
mkdir -p $LOCAL_WORK_DPATH
cd $LOCAL_WORK_DPATH

docker run \
    --runtime=nvidia \
    --volume "$LOCAL_WORK_DPATH":/tmp/ingress \
    --volume $HOME/.aws:/root/.aws:ro \
    --volume "$HOME/code":/extern_code:ro \
    --volume "$HOME/data":/extern_data:ro \
    --volume "$HOME"/.cache/pip:/pip_cache \
    --env AWS_PROFILE=iarpa \
    -it registry.smartgitlab.com/kitware/geowatch:0.16.2-17a6765ff-strict-pyenv3.11.2-20240410T192350-0400-from-0da55667 bash

ipython
from geowatch.cli.smartflow.run_sc_fusion import *  # NOQA

config = SCFusionConfig(**{
    'input_path'                    : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/split_work/52SDG98/products/ac_datagen/items.jsonl',
    'input_region_path'             : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/split_input/52SDG98/region_models/KR_R002.geojson',
    'output_path'                   : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/split_work/52SDG98/products/sc-fusion/items.jsonl',
    'aws_profile'                   : None,
    'dryrun'                        : False,
    'outbucket'                     : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval21_batch_v193/batch/kit/KR_R002/split_work/52SDG98/products/sc-fusion',
    'ta2_s3_collation_bucket'       : None,
    'sc_pxl_config'                 : 'batch_size: 1\nchip_dims: auto\nchip_overlap: 0.3\ndrop_unused_frames: true\ninput_space_scale: 4GSD\nmask_low_quality: true\nnum_workers: 8\nobservable_threshold: 0.0\noutput_space_scale: 4GSD\npackage_fpath: /root/data/smart_expt_dvc/models/fusion/Drop7-Cropped2GSD/packages/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84_epoch17_step1548.pt\nresample_invalid_frames: 3\nset_cover_algo: null\ntta_fliprot: 0.0\ntta_time: 0.0\nwindow_space_scale: 4GSD\nwrite_workers: 0',
    'sc_poly_config'                : 'boundaries_as: bounds\nmin_area_square_meters: 7200\nnew_algo: crall\npolygon_simplify_tolerance: 1\nresolution: 8GSD\nsite_score_thresh: 0.3\nsmoothing: 0.0\nthresh: 0.3',
    'input_region_models_asset_name': 'sv_out_region_models',
    'input_site_models_asset_name'  : 'sv_out_site_models',
    'egress_intermediate_outputs'   : 1,
})

"""


class SCFusionConfig(scfg.DataConfig):
    """
    Run TA-2 SC fusion as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='DEPRECATED. DO NOT USE')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))

    ta2_s3_collation_bucket = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        S3 Location for collated TA-2 output (bucket name should
        include up to eval name)
        '''))

    sc_pxl_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for fusion.predict.
        '''))

    sc_poly_config = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Raw json/yaml or a path to a json/yaml file that specifies the
        config for SC tracking.
        '''))

    input_region_models_asset_name = scfg.Value('sv_out_region_models', type=str, required=False, help=ub.paragraph(
        '''
        Which region model assets to use as input
        '''))

    input_site_models_asset_name = scfg.Value('sv_out_site_models', type=str, required=False, help=ub.paragraph(
        '''
        Which site model assets to to use as input
        '''))

    egress_intermediate_outputs = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        If true egress intermediate heatmaps, otherwise only egress the geojson
        '''))


def main(cmdline=1, **kwargs):
    config = SCFusionConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    print('config = {}'.format(ub.urepr(config, nl=1, align=':')))
    run_sc_fusion_for_baseline(config)


def run_sc_fusion_for_baseline(config):
    from geowatch.cli.smartflow_ingress import smartflow_ingress
    from geowatch.cli.smartflow_egress import smartflow_egress
    from geowatch.tasks.fusion.predict import predict  # NOQA
    from geowatch.tasks.fusion.datamodules.temporal_sampling import TimeSampleError
    from geowatch.utils.util_framework import download_region, determine_region_id
    from kwutil.util_yaml import Yaml
    from geowatch.utils import util_framework
    from geowatch.mlops import smart_pipeline
    import kwcoco

    if config.aws_profile is not None:
        # This should be sufficient, but it is not tested.
        # TODO: use the new register_bucket API.
        from geowatch.utils import util_fsspec
        util_fsspec.S3Path._new_fs(profile=config.aws_profile)

    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")

    # ingress_dir = ub.Path('/home/joncrall/data/dvc-repos/smart_expt_dvc/_airflow/temp').ensuredir()
    ingress_dir = ub.Path('/tmp/ingress')

    # input_region_asset_name = 'sv_out_region_models'
    input_region_asset_name = config.input_region_models_asset_name

    ingressed_assets = smartflow_ingress(
        input_path=config.input_path,
        assets=[
            # {'key': 'cropped_region_models_bas'},
            # {'key': 'sv_out_region_models', 'allow_missing': False},

            {'key': input_region_asset_name, 'allow_missing': False},

            # {'key': 'cropped_kwcoco_for_sc'},
            # {'key': 'cropped_kwcoco_for_sc_assets'}
            {'key': 'enriched_acsc_kwcoco_file'},
            {'key': 'enriched_acsc_kwcoco_teamfeats'},
            {'key': 'enriched_acsc_kwcoco_rawbands'},
        ],
        outdir=ingress_dir,
        aws_profile=config.aws_profile,
        dryrun=config.dryrun
    )

    input_site_summary_dpath = ingressed_assets[input_region_asset_name]
    assert os.path.exists(input_site_summary_dpath)
    print(f'Found input site summary dpath: {input_site_summary_dpath}')
    ub.cmd(f'geowatch site_stats "{input_site_summary_dpath}/*.geojson"', verbose=3)

    # # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'

    download_region(
        input_region_path=config.input_region_path,
        output_region_path=local_region_path,
        aws_profile=config.aws_profile,
        strip_nonregions=True,
    )

    # Determine the region_id in the region file.
    region_id = determine_region_id(local_region_path)

    sc_fusion_kwcoco_path = ingress_dir / 'sc_fusion_kwcoco.json'
    tracked_sc_kwcoco_path = sc_fusion_kwcoco_path.augment(stemsuffix='_tracked')

    site_models_outdir = (ingress_dir / 'sc_out_site_models').ensuredir()
    region_models_outdir = (ingress_dir / 'sc_out_region_models').ensuredir()
    site_models_manifest_outdir = (ingress_dir / 'tracking_manifests_sc').ensuredir()

    site_models_manifest_outpath = site_models_manifest_outdir / 'site_models_manifest.json'
    # Copy input region model into region_models outdir to be updated
    # (rather than generated from tracking, which may not have the
    # same bounds as the original)
    shutil.copy(local_region_path, region_models_outdir / f'{region_id}.geojson')

    region_models_manifest_fpath = ingress_dir / 'sc_out_region_models_manifest.json'

    node_state.print_current_state(ingress_dir)

    # 3.1. Check that we have at least one "video" (BAS identified
    # site) to run over; if not skip SC fusion and KWCOCO to GeoJSON

    # TODO: could use kwcoco info to get lazy loading of just the header.
    input_kwcoco_fpath = ingressed_assets['enriched_acsc_kwcoco_file']
    ingress_dset = kwcoco.CocoDataset(input_kwcoco_fpath, autobuild=False)
    if ingress_dset.n_videos > 0:
        # 3. Run fusion
        print('*********************')
        print("* Running SC fusion *")

        ub.cmd(f'kwcoco stats {input_kwcoco_fpath}', verbose=3)
        ub.cmd(f'geowatch stats {input_kwcoco_fpath}', verbose=3)

        # The params should be fully given in the DAG.
        sc_pxl_config = Yaml.coerce(config.sc_pxl_config or {})
        if sc_pxl_config.get('package_fpath', None) is None:
            raise ValueError('Requires package_fpath')

        # Debugging
        # foo = '/root/data/smart_expt_dvc/models/fusion/Drop7-Cropped2GSD/packages/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84/Drop7-Cropped2GSD_SC_bgrn_gnt_split6_V84_epoch17_step1548.pt'
        # sc_pxl_config['package_fpath'] = foo
        # sc_pxl_config['set_cover_algo'] = 'approx'

        sc_pxl = smart_pipeline.SC_HeatmapPrediction(root_dpath=ingress_dir)
        sc_pxl.configure({
            'pred_pxl_fpath': sc_fusion_kwcoco_path,
            'test_dataset': ingressed_assets['enriched_acsc_kwcoco_file'],
        } | sc_pxl_config)

        if False:
            try:
                # Quick inspection of how big the predicted videos are going to be.
                from geowatch.utils import util_resolution
                from geowatch.utils import util_units
                import numpy as np
                # from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDatasetConfig
                # sc_pxl_config['chip_dims']
                # input_resolution = util_resolution.ResolvedUnit.coerce(sc_pxl_config['input_space_scale'])
                num_output_channels = 6  # hard coded for the quick check
                for video in ingress_dset.dataset['videos']:
                    output_resolution = util_resolution.ResolvedUnit.coerce(sc_pxl_config['output_space_scale'])
                    output_resolution = util_resolution.ResolvedUnit.coerce('4GSD')
                    vidname = video['name']
                    vid_width = video['width']
                    vid_height = video['height']
                    num_frames = video['num_frames']
                    video_resolution = util_resolution.ResolvedUnit.coerce(video['resolution'])
                    video_dsize = util_resolution.ResolvedWindow((vid_width, vid_height), video_resolution)
                    # input_dsize = video_dsize.at_resolution(input_resolution)
                    output_dsize = video_dsize.at_resolution(output_resolution)
                    output_frame_pixels = np.prod(output_dsize.window)
                    bytes_per_cell = np.dtype('float32').itemsize
                    num_raster_cells = output_frame_pixels * num_frames
                    total_bytes = num_raster_cells * bytes_per_cell
                    ureg = util_units.unit_registry()
                    size_per_channel = (total_bytes * ureg.bytes).to('gigabytes')
                    total_size = num_output_channels * size_per_channel
                    print(f'{vidname} - {total_size:0.2f} - WH={output_dsize},  T={num_frames}, C={num_output_channels}')

                    memory_max_gb = 8
                    factor = np.sqrt(total_size.m / memory_max_gb)
                    if factor > 1:
                        output_resolution = util_resolution.ResolvedUnit(output_resolution.mag * factor, output_resolution.unit)
                        output_dsize = video_dsize.at_resolution(output_resolution)
                        output_frame_pixels = np.prod(output_dsize.window)
                        bytes_per_cell = np.dtype('float32').itemsize
                        num_raster_cells = output_frame_pixels * num_frames
                        total_bytes = num_raster_cells * bytes_per_cell
                        ureg = util_units.unit_registry()
                        size_per_channel = (total_bytes * ureg.bytes).to('gigabytes')
                        total_size = num_output_channels * size_per_channel
                        # output_resolution = util_resolution.ResolvedUnit.coerce(sc_pxl_config['output_space_scale'])
                    print(f'{vidname} - {total_size:0.2f} - WH={output_dsize},  T={num_frames}, C={num_output_channels}')

                ingress_dset.videos()
            except Exception as ex:
                print('Quick Introspection failed:')
                print(f'ex = {ub.urepr(ex, nl=1)}')

        command = sc_pxl.command()
        print(command)
        try:
            ub.cmd(command, check=True, verbose=3, system=True)
            node_state.print_current_state(ingress_dir)
            # ub.cmd(['kwcoco', 'reroot', f'--src={sc_fusion_kwcoco_path}', '--inplace=1', '--absolute=0'])
            # ub.cmd(['geowatch', 'visualize', f'{sc_fusion_kwcoco_path}', '--smart'], verbose=3)
        except TimeSampleError:
            # FIXME: wont work anymore with mlops. Not sure if needed.
            # Can always catch a CalledProcessError and inspect stdout
            print("* Error with time sampling during SC Predict "
                  "(shown below) -- attempting to continue anyway")
            traceback.print_exception(*sys.exc_info())
        else:
            ub.cmd(f'kwcoco stats {sc_fusion_kwcoco_path}', verbose=3)
            ub.cmd(f'geowatch stats {sc_fusion_kwcoco_path}', verbose=3)

            # 4. Compute tracks (SC)
            print('*************************')
            print("* Computing tracks (SC) *")

            # Params are fully specified in the DAG
            sc_track_kwargs = Yaml.coerce(config.sc_poly_config or {})
            final_sc_poly_config = {
                'pred_pxl_fpath': sc_fusion_kwcoco_path,               # Sets --input_kwcoco
                'site_summaries_fpath': region_models_manifest_fpath,  # Sets --out_site_summaries_fpath
                'site_summaries_dpath': region_models_outdir,          # Sets --out_site_summaries_dir
                'sites_dpath': site_models_outdir,                     # Sets --out_sites_dir
                'sites_fpath': site_models_manifest_outpath,           # Sets --out_sites_fpath
                'poly_kwcoco_fpath': tracked_sc_kwcoco_path,           # Sets --out_kwcoco
                'site_summary': ub.Path(input_site_summary_dpath) / '*.geojson',  # Sets --site_summary
                'append_mode': True,
            } | sc_track_kwargs
            # See: ~/code/watch/geowatch/mlops/smart_pipeline.py
            sc_poly = smart_pipeline.SC_PolygonPrediction(root_dpath=ingress_dir)
            sc_poly.configure(final_sc_poly_config)
            command = sc_poly.command()
            ub.cmd(command, check=True, verbose=3, system=True)

            node_state.print_current_state(ingress_dir)

            ub.cmd(f'kwcoco stats {tracked_sc_kwcoco_path}', verbose=3)
            ub.cmd(f'geowatch stats {tracked_sc_kwcoco_path}', verbose=3)
    else:
        print('Warning: No Videos in Ingress Dataset, Skipping Predict!')

    cropped_site_models_outdir = ingress_dir / 'cropped_site_models_sc'
    cropped_region_models_outdir = ingress_dir / 'cropped_region_models_sc'
    sc_heatmap_dpath = (ingress_dir / '_assets').ensuredir()

    cropped_site_models_outdir.ensuredir()
    cropped_region_models_outdir.ensuredir()

    ub.cmd([
        'python', '-m', 'geowatch.cli.crop_sites_to_regions',
        '--site_models', site_models_outdir / '*.geojson',
        '--region_models', region_models_outdir / f'{region_id}.geojson',
        '--new_site_dpath', cropped_site_models_outdir,
        '--new_region_dpath', cropped_region_models_outdir
    ], check=True, verbose=3, capture=False)

    node_state.print_current_state(ingress_dir)

    # Validate and fix all outputs
    print('Fixup and validate outputs')
    util_framework.fixup_and_validate_site_and_region_models(
        region_dpath=cropped_region_models_outdir,
        site_dpath=cropped_site_models_outdir,
    )

    node_state.print_current_state(ingress_dir)
    node_state.print_directory_contents(cropped_site_models_outdir)
    node_state.print_directory_contents(cropped_region_models_outdir)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    ingressed_assets['cropped_site_models_sc'] = cropped_site_models_outdir
    ingressed_assets['cropped_region_models_sc'] = cropped_region_models_outdir

    # Add in intermediate outputs for debugging
    EGRESS_INTERMEDIATE_OUTPUTS = config.egress_intermediate_outputs
    if EGRESS_INTERMEDIATE_OUTPUTS:
        # Reroot kwcoco files to make downloaded results easier to work with
        if sc_fusion_kwcoco_path.exists():
            ub.cmd(['kwcoco', 'reroot', f'--src={sc_fusion_kwcoco_path}', '--inplace=1', '--absolute=0'])
            ingressed_assets['sc_heatmap_kwcoco_file'] = sc_fusion_kwcoco_path
        if tracked_sc_kwcoco_path.exists():
            ub.cmd(['kwcoco', 'reroot', f'--src={tracked_sc_kwcoco_path}', '--inplace=1', '--absolute=0'])
            ingressed_assets['sc_tracked_kwcoco_file'] = tracked_sc_kwcoco_path
        ingressed_assets['sc_heatmap_assets'] = sc_heatmap_dpath
        ingressed_assets['sc_tracking_manifest_dpath'] = site_models_manifest_outdir
        if region_models_manifest_fpath.exists():
            ingressed_assets['sc_tracking_manifest_fpath'] = region_models_manifest_fpath

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=config.aws_profile,
                     dryrun=False,
                     newline=False)

    # 6. (Optional) collate TA-2 output
    if config.ta2_s3_collation_bucket is not None:
        print("* Collating TA-2 output")
        util_framework.ta2_collate_output(None,
                                          cropped_region_models_outdir,
                                          cropped_site_models_outdir,
                                          config.ta2_s3_collation_bucket)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class FixupPredictKwcocoMetadata(scfg.DataConfig):
    """
    Update pre-0.15.1 kwcoco predictions to properly store train-time params.

    The kwcoco info section of kwcoco files produced by geowatch.fusion.predict
    only contained the "data" section of the fit configuration. This script is
    able to fix one or more of those old predicted files as long as the path to
    the model is available. Warnings that direct users to this help document
    will typically give example usage that fixes one file, but multiple files
    can be fixed at once by specifying a glob pattern. The following example
    illustrates this.

    CommandLine:
        # Say you have an old mlops directory of results
        DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase3_expt' --hardware=auto)
        MLOPS_DPATH=$DVC_EXPT_DPATH/_preeval20_bas_grid

        # Construct a glob pattern that matches the kwcoco files that need to
        # be fixed, and pass it to this script.
        python -m geowatch.cli.experimental.fixup_predict_kwcoco_metadata \\
            --coco_fpaths "$MLOPS_DPATH/pred/flat/bas_pxl/*/pred.kwcoco.zip"

    Note:
        This modifies input files INPLACE!
    """
    coco_fpaths = scfg.Value(None, help='Path to one or more predicted kwcoco files to fix')
    workers = scfg.Value(8)

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> from fixup_bad_fit_config import *  # NOQA
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> cls = FixupBadFitConfigCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import rich
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        rich.print('config = ' + ub.urepr(config, nl=1))

        from kwutil import util_path
        from kwutil import util_progress

        kwcoco_fpaths = util_path.coerce_patterned_paths(config.coco_fpaths)
        pman = util_progress.ProgressManager()
        jobs = ub.JobPool(mode='process', max_workers=config.workers)
        with pman, jobs:
            for node_dpath in pman.progiter(kwcoco_fpaths, desc='submit kwcoco fixup jobs'):
                jobs.submit(fixup_pxl_pred_node_dpath, node_dpath)

            stats = ub.ddict(int)
            for job in pman.progiter(jobs.as_completed(), desc='collect fixup jobs'):
                status = job.result()
                stats[status['message']] += 1
                pman.update_info(f'stats = {ub.urepr(stats, nl=1)}')


def fixup_pxl_pred_node_dpath(coco_fpath):
    import kwcoco
    from kwutil.util_yaml import Yaml
    import zipfile
    status = {}
    messages = []
    dset = kwcoco.CocoDataset(coco_fpath)
    candidates = []
    for item in dset.dataset['info']:
        if item['type'] == 'process':
            if item['properties']['name'] == 'geowatch.tasks.fusion.predict':
                candidates.append(item)
    assert len(candidates) == 1
    pred_item = candidates[0]
    extra = pred_item['properties']['extra']
    if 'fit_config' in extra:
        if set(extra['fit_config']).issuperset({'data', 'trainer'}):
            messages.append('Already had updated fit config')
        else:
            status['message'] = 'Updated metadata'
            package_fpath = pred_item['properties']['config']['package_fpath']
            zfile = zipfile.ZipFile(package_fpath)
            fit_config = None
            with zfile:
                for name in zfile.namelist():
                    if name.endswith('package_header/config.yaml'):
                        config_text = zfile.read(name).decode('utf8')
                        fit_config = Yaml.loads(config_text, backend='pyyaml')
                        break
            assert fit_config is not None
            extra['fit_config'] = fit_config
            dset.dump()
            messages.append('Updated metadata')
    status['message'] = '\n'.join(messages)
    return status

__cli__ = FixupPredictKwcocoMetadata
main = __cli__.main

if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/geowatch/dev/oneoffs/fixup_predict_kwcoco_metadata.py
    """
    main()



#!/usr/bin/env python3
import ubelt as ub
import scriptconfig as scfg
from geowatch.cli.smartflow_ingress import smartflow_ingress
from geowatch.cli.smartflow_egress import smartflow_egress


class TeamFeatInvariantsConfig(scfg.DataConfig):
    """
    Run UKY invariant feature computation as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))
    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')
    model_path = scfg.Value(None, type=str, required=True, help='File path to UKY invariants model')
    pca_projection_path = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            File path to UKY invariants PCA projections file
            '''))
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    do_pca = scfg.Value(False, isflag=True, help='Perform PCA on invariants')
    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
            '''
            Output as simple newline separated STAC items
            '''))


def main():
    config = TeamFeatInvariantsConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_uky_invariants_for_baseline(config)


def run_uky_invariants_for_baseline(config):
    from geowatch.utils.util_framework import download_region

    ####
    # DEBUGGING:
    # Print info about what version of the code we are running on
    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')

    ingressed_assets = smartflow_ingress(
        config.input_path,
        ['enriched_bas_kwcoco_file',
         'enriched_bas_kwcoco_teamfeats',
         'enriched_bas_kwcoco_rawbands'],
        ingress_dir,
        config.aws_profile,
        config.dryrun)

    timecombined_kwcoco_file_for_bas = ingressed_assets['enriched_bas_kwcoco_file']

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = ub.Path('/tmp/region.json')
    download_region(
        input_region_path=config.input_region_path,
        output_region_path=local_region_path,
        aws_profile=config.aws_profile,
        strip_nonregions=True,
    )

    # 2. Generate Invariants
    print("* Generating UKY/WU invariant features *")
    # FIXME: probably should not be called "SC" invariants
    invariants_kwcoco_path = ingress_dir / 'sc_invariants_kwcoco.json'
    ub.cmd([
        'python', '-m', 'geowatch.tasks.invariants.predict',
        '--input_kwcoco', timecombined_kwcoco_file_for_bas,
        '--output_kwcoco', invariants_kwcoco_path,
        '--pretext_package_path', config.model_path,
        '--pca_projection_path', config.pca_projection_path,
        '--assets_dname', '_teamfeats',
        '--input_space_scale', '30GSD',
        '--window_space_scale', '30GSD',
        '--patch_size', '256',
        '--do_pca', str(1 if config.do_pca else 0),
        '--patch_overlap', '0.3',
        '--num_workers', '2',
        '--write_workers', '0',
        '--tasks', 'before_after', 'pretext'
    ], check=True, verbose=3, capture=False)

    # 3. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")

    # Reroot kwcoco files to make downloaded results easier to work with
    ub.cmd(['kwcoco', 'reroot', f'--src={invariants_kwcoco_path}', '--inplace=1', '--absolute=0'])

    ingressed_assets['enriched_bas_kwcoco_file'] = invariants_kwcoco_path
    ingressed_assets['enriched_bas_kwcoco_teamfeats'] = ingress_dir / '_teamfeats'

    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=None,
                     dryrun=False,
                     newline=False)


if __name__ == "__main__":
    main()




def test_aggregate_with_unhashable_columns():
    from geowatch.mlops.aggregate import Aggregator
    agg = Aggregator.demo(rng=0, num=100, include_unhashable=True)
    agg.build()
    assert isinstance(agg.params['params.demo_node.unhashable_list'].iloc[0], list), 'The raw params should remember the object form if possible'
    assert isinstance(agg.effective_params['params.demo_node.unhashable_list'].iloc[0], str), 'The effective params should all be hashable'
    assert isinstance(agg.params['params.demo_node.unhashable_dict'].iloc[0], dict), 'The raw params should remember the object form if possible'
    assert isinstance(agg.effective_params['params.demo_node.unhashable_dict'].iloc[0], str), 'The effective params should all be hashable'



__mkinit__ = """
mkinit ~/code/watch/geowatch/utils/__init__.py --lazy --diff
mkinit ~/code/watch/geowatch/utils/__init__.py --lazy -w

TEST:
    python -c "from geowatch import utils"
    EAGER_IMPORT_MODULES=geowatch python -c "from geowatch import utils"
"""

__submodules__ = {
    '*': [],
    'util_data': ['find_smart_dvc_dpath', 'find_dvc_dpath'],
    # 'util_yaml': ['Yaml'],
}


def lazy_import(module_name, submodules, submod_attrs, eager='auto'):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    eager_import_flag = False
    if eager == 'auto':
        eager_import_text = os.environ.get('EAGER_IMPORT', '')
        if eager_import_text:
            eager_import_text_ = eager_import_text.lower()
            if eager_import_text_ in {'true', '1', 'on', 'yes'}:
                eager_import_flag = True

        eager_import_module_text = os.environ.get('EAGER_IMPORT_MODULES', '')
        if eager_import_module_text:
            if eager_import_module_text.lower() in __name__.lower():
                eager_import_flag = True
    else:
        eager_import_flag = eager

    if eager_import_flag:
        for name in submodules:
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'configargparse_ext',
        'ext_monai',
        'ijson_ext',
        'kwcoco_extensions',
        'lightning_ext',
        'process_context',
        'remedian',
        'result_analysis',
        'reverse_hashid',
        'simple_dvc',
        'util_bands',
        'util_codes',
        'util_data',
        'util_dotdict',
        'util_exception',
        'util_framework',
        'util_fsspec',
        'util_gdal',
        'util_girder',
        'util_gis',
        'util_globals',
        'util_hardware',
        'util_iter',
        'util_kwarray',
        'util_kwimage',
        'util_kwplot',
        'util_logging',
        'util_nesting',
        'util_netharn',
        'util_nvidia',
        'util_pandas',
        'util_param_grid',
        'util_raster',
        'util_regex',
        'util_resolution',
        'util_retry',
        'util_s3',
        'util_stringalgo',
        'util_torchmetrics',
        'util_units',
    },
    submod_attrs={
        'util_data': [
            'find_smart_dvc_dpath',
            'find_dvc_dpath',
        ],
    },
)


def __dir__():
    return __all__

__all__ = ['configargparse_ext', 'ext_monai', 'find_dvc_dpath',
           'find_smart_dvc_dpath', 'ijson_ext', 'kwcoco_extensions',
           'lightning_ext', 'process_context', 'remedian', 'result_analysis',
           'reverse_hashid', 'simple_dvc', 'util_bands',
           'util_codes', 'util_data', 'util_dotdict', 'util_exception',
           'util_framework', 'util_fsspec', 'util_gdal', 'util_girder',
           'util_gis', 'util_globals', 'util_hardware', 'util_iter',
           'util_kwarray', 'util_kwimage', 'util_kwplot', 'util_logging',
           'util_nesting', 'util_netharn', 'util_nvidia', 'util_pandas',
           'util_param_grid', 'util_raster', 'util_regex', 'util_resolution',
           'util_retry', 'util_s3', 'util_stringalgo',
           'util_torchmetrics', 'util_units']





def mwe_load_spacetime_region():
    import kwimage
    # import kwcoco
    import numpy as np
    from watch.demo import coerce_kwcoco
    from typing import List
    coco_dset = coerce_kwcoco('watch-msi', dates=True, geodata=True)

    # Keep consistent bands
    keep_ids = [img.img['id'] for img in coco_dset.images().coco_images if 'B11' in img.channels]
    coco_dset = coco_dset.subset(keep_ids)

    import ndsampler
    # from watch.utils import util_kwimage
    sampler = ndsampler.CocoSampler(coco_dset)

    gid1, gid2, gid3 = coco_dset.videos().images[0][0:3]

    # Choose channels likely to be part of the demo set
    chans = 'B11'

    target = {
        'space_slice': kwimage.Boxes([[20, 30, 301, 307]], format='xywh').to_slices()[0],
        'gids': [gid1, gid2, gid3],
        'channels': chans,
    }

    sample = sampler.load_sample(target, with_annots=True)

    detections: List[kwimage.Detections] = sample['annots']['frame_dets']
    imdata = sample['im']

    frame_canvases = []

    for frame, dets in zip(imdata, detections):
        canvas = kwimage.normalize_intensity(frame)
        # fill_nans_with_checkers seems to not work with multi channels
        # this is a hack to fix it.
        canvas = np.concatenate([kwimage.fill_nans_with_checkers(c) for c in canvas.transpose(2, 0, 1)], axis=2)
        assert not np.isnan(canvas).sum()
        canvas = dets.draw_on(canvas)
        frame_canvases.append(canvas)

    final_canvas = kwimage.stack_images(frame_canvases)
    import kwplot
    kwplot.autompl()
    kwplot.imshow(final_canvas)



"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/plots.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import show_affinity_sample_process  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import plot_dense_sample_indices  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import plot_temporal_sample_indices  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import plot_temporal_sample  # NOQA



#!/usr/bin/env python3
import sys
import os
import scriptconfig as scfg
import ubelt as ub


class BaselineFrameworkIngressConfig(scfg.DataConfig):
    """
    Ingress data from T&E baseline framework input file. The output will be stored as a json catalog
    """
    input_path = scfg.Value(None, type=str, position=1, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    outdir = scfg.Value(None, type=str, short_alias=['o'], help=ub.paragraph(
            '''
            Output directory for ingressed assets an output STAC Catalog
            '''))
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    show_progress = scfg.Value(False, isflag=True, short_alias=['s'], help='Show progress for AWS CLI commands')
    requester_pays = scfg.Value(False, isflag=True, short_alias=['r'], help=ub.paragraph(
            '''
            Run AWS CLI commands with `--requestor_payer requester` flag
            '''))
    jobs = scfg.Value(1, type=str, short_alias=['j'], help='Number of jobs to run in parallel')
    virtual = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Replace asset hrefs with GDAL Virtual File System links
            '''))
    catalog_fpath = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Name of the ouptut catalog.
            Defaults to <outdir>/catalog.json
            '''))
    relative = scfg.Value(False, isflag=True, help='if true use relative paths')

    def __post_init__(self):
        # super().__post_init__()
        if self.catalog_fpath is None and self.outdir is not None:
            self.catalog_fpath = os.path.join(self.outdir, 'catalog.json')
        if self.catalog_fpath is not None and self.outdir is None:
            self.outdir = os.path.dirname(os.path.abspath(self.catalog_fpath))


def main():
    config = BaselineFrameworkIngressConfig.cli(strict=True)
    import rich
    rich.print(ub.urepr(config))

    if config.outdir is None:
        raise ValueError('outdir is required')

    if config.input_path is None:
        raise ValueError('input_path is required')

    baseline_framework_ingress(**config)


def baseline_framework_ingress(input_path,
                               outdir,
                               catalog_fpath=None,
                               aws_profile=None,
                               dryrun=False,
                               show_progress=False,
                               requester_pays=False,
                               relative=False,
                               jobs=1,
                               virtual=False):

    from kwutil import util_parallel
    from kwutil import util_progress
    import rich
    import pystac
    import traceback
    from geowatch.utils import util_fsspec
    from geowatch.utils.util_framework import ingress_item

    workers = util_parallel.coerce_num_workers(jobs)
    print(f'Runing baseline_framework_ingress with workers={workers}')

    os.makedirs(outdir, exist_ok=True)

    if relative:
        catalog_type = pystac.CatalogType.RELATIVE_PUBLISHED
    else:
        catalog_type = pystac.CatalogType.ABSOLUTE_PUBLISHED

    if catalog_fpath is None:
        catalog_fpath = os.path.join(outdir, 'catalog.json')
    catalog = pystac.Catalog('Baseline Framework ingress catalog',
                             'STAC catalog of SMART search results',
                             href=catalog_fpath, catalog_type=catalog_type)

    catalog.set_root(catalog)

    if relative:
        catalog.make_all_asset_hrefs_relative()

    if aws_profile is not None:
        aws_base_command = ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    if dryrun:
        aws_base_command.append('--dryrun')

    if not show_progress:
        aws_base_command.append('--no-progress')

    if requester_pays:
        aws_base_command.extend(['--request-payer', 'requester'])

    if aws_profile is not None or requester_pays:
        # This should be sufficient, but it is not tested.
        util_fsspec.S3Path._new_fs(
            profile=aws_profile, requester_pays=requester_pays)

    input_stac_items = load_input_stac_items(input_path, aws_base_command)

    print(f'Loaded {len(input_stac_items)} stac items')

    ingress_kw = {
        'outdir': outdir,
        'aws_base_command': aws_base_command,
        'dryrun': dryrun,
        'relative': relative,
        'virtual': virtual,
    }

    pool = ub.JobPool(mode='thread' if workers > 1 else 'serial',
                      max_workers=workers)
    pman = util_progress.ProgressManager(backend='rich')
    with pman:
        """
        DEVELOPER NOTE:
            There is something that can cause a lockup here. To reproduce
            first ensure that the outdir is cleared, so no caching happens.
            The failure seems to happen when the mode is process. Using thread
            or serial seems fine.

            Update: the issue seems to happen if you use the pool.__enter__
            method before pman. Using it after seems ok with progiter, but not
            rich. Removing the __enter__ does not help the rich case, and now
            switching back to progiter, the lockup is happening again...
        """
        for feature in pman.progiter(input_stac_items, desc='submit ingress jobs'):
            pool.submit(ingress_item, feature, **ingress_kw)

        for job in pman.progiter(pool.as_completed(), total=len(pool), desc='ingress items'):
            try:
                mapped_item = job.result()
            except Exception:
                rich.print("[yellow]WARNING: Exception occurred (printed below), dropping item!")
                traceback.print_exception(*sys.exc_info())
                continue
            else:
                # print(mapped_item.to_dict())
                catalog.add_item(mapped_item)
    print('Finished downloads, saving catalog')
    catalog.save(catalog_type=catalog_type)
    print('wrote catalog_fpath = {!r}'.format(catalog_fpath))
    return catalog


def read_input_stac_items(path):
    """
    Read the stac input format from a file on disk.

    This also handles jsonl files as well as a a fallback for whitespace
    separated data.

    Args:
        path (str | PathLike): path to the .input file

    Returns:
        List[Dict]: stac items
    """
    import json

    def _open(p, m):
        # handle fsspath objects gracefully
        if hasattr(path, 'open'):
            return path.open(m)
        else:
            return open(path, m)
    try:
        with _open(path, 'r') as f:
            input_json = json.load(f)
        items = input_json['stac'].get('features', [])
    # Excepting KeyError here in case of a single line STAC item input
    except (json.decoder.JSONDecodeError, KeyError):
        try:
            # Support for simple newline separated STAC items
            with _open(path, 'r') as f:
                items = [json.loads(line) for line in f]
        except json.decoder.JSONDecodeError:
            # Support for whitespace separated data
            with _open(path, 'r') as f:
                text = f.read()
            items = []
            stack = [line for line in text.split('\n')[::-1] if line]
            while stack:
                line = stack.pop()
                try:
                    item = json.loads(line)
                except json.decoder.JSONDecodeError as e:
                    # Hack for the case where a new line is missing
                    if line[e.pos] == '{':
                        stack.append(line[e.pos:].strip())
                        stack.append(line[:e.pos])
                    else:
                        raise
                else:
                    items.append(item)
    return items


def load_input_stac_items(input_path, aws_base_command):
    """
    Load the stac input format from a file on disk or AWS
    """
    import tempfile

    if aws_base_command is None or not input_path.startswith('s3'):
        # New method
        from geowatch.utils import util_fsspec
        input_path = util_fsspec.FSPath.coerce(input_path)
        input_stac_items = read_input_stac_items(input_path)
    else:
        # Old method, remove once we can
        with tempfile.NamedTemporaryFile() as temporary_file:
            ub.cmd(
                [*aws_base_command, input_path, temporary_file.name],
                check=True, verbose=3)

            input_stac_items = read_input_stac_items(temporary_file.name)

    return input_stac_items


if __name__ == '__main__':
    main()



import numpy as np


def get_seg_display(seg):
    seg_display = np.zeros([seg.shape[0], seg.shape[1], 4], dtype=np.float)
    if len(seg.shape) == 2:
        seg_display[..., 0] = seg
        seg_display[..., 3] = seg
    else:
        for i in range(seg.shape[-1]):
            seg_display[..., i] = seg[..., i]
        seg_display[..., 3] = np.clip(np.sum(seg, axis=-1), 0, 1)
    return seg_display


def batch_to_cuda(batch):
    # Send data to computing device:
    for key, item in batch.items():
        if hasattr(item, "cuda"):
            batch[key] = item.cuda(non_blocking=True)
    return batch


def batch_to_cpu(batch):
    # Send data to computing device:
    for key, item in batch.items():
        if hasattr(item, "cpu"):
            batch[key] = item.cpu()
    return batch



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains Axial-ResNet model instances for Axial-DeepLab and MaX-DeepLab.

Reference:
    Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
      ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
        Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
          Liang-Chieh Chen.
    MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
      CVPR 2021. https://arxiv.org/abs/2012.00759
        Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
"""

import abc
import collections.abc
import copy

import tensorflow as tf
from absl import logging
from deeplab2.model.encoder import axial_resnet


def _get_default_config():
    """Gets the default config for Axial-ResNets."""
    # The default config dictionary for an Axial-ResNet is the MaX-DeepLab-S
    # architecture for panoptic segmentation. This default config dictionary also
    # exactly matches the default arguments of the functions.
    default_config = {
        'num_blocks': [3, 4, 6, 3],
        'backbone_layer_multiplier': 1.0,
        'width_multiplier': 1.0,
        'stem_width_multiplier': 1.0,
        'output_stride': 16,
        'classification_mode': False,
        'backbone_type': 'resnet_beta',
        'use_axial_beyond_stride': 16,
        'backbone_use_transformer_beyond_stride': 32,
        'extra_decoder_use_transformer_beyond_stride': 32,
        'backbone_decoder_num_stacks': 0,
        'backbone_decoder_blocks_per_stage': 1,
        'extra_decoder_num_stacks': 0,
        'extra_decoder_blocks_per_stage': 1,
        'max_num_mask_slots': 128,
        'num_mask_slots': 128,
        'memory_channels': 256,
        'base_transformer_expansion': 1.0,
        'global_feed_forward_network_channels': 256,
        'high_resolution_output_stride': 4,
        'activation': 'relu',
        'block_group_config': {
            'attention_bottleneck_expansion': 2,
            'drop_path_keep_prob': 0.8,
            'drop_path_beyond_stride': 16,
            'drop_path_schedule': 'constant',
            'positional_encoding_type': None,
            'use_global_beyond_stride': 0,
            'use_sac_beyond_stride': 0,
            'use_squeeze_and_excite': False,
            'conv_use_recompute_grad': False,
            'axial_use_recompute_grad': True,
            'recompute_within_stride': 0,
            'transformer_use_recompute_grad': False,
            'axial_layer_config': {
                'query_shape': (129, 129),
                'key_expansion': 1,
                'value_expansion': 2,
                'memory_flange': (32, 32),
                'double_global_attention': False,
                'num_heads': 8,
                'use_query_rpe_similarity': True,
                'use_key_rpe_similarity': True,
                'use_content_similarity': True,
                'retrieve_value_rpe': True,
                'retrieve_value_content': True,
                'initialization_std_for_query_key_rpe': 1.0,
                'initialization_std_for_value_rpe': 1.0,
                'self_attention_activation': 'softmax',
            },
            'dual_path_transformer_layer_config': {
                'num_heads': 8,
                'bottleneck_expansion': 2,
                'key_expansion': 1,
                'value_expansion': 2,
                'feed_forward_network_channels': 2048,
                'use_memory_self_attention': True,
                'use_pixel2memory_feedback_attention': True,
                'transformer_activation': 'softmax',
            },
        },
        'bn_layer': tf.keras.layers.BatchNormalization,
        'conv_kernel_weight_decay': 0.0,
    }
    return default_config


def override(config_dict, override_dict):
    """Recursively overrides a config dict with another."""
    output_dict = copy.deepcopy(config_dict)
    for key, value in override_dict.items():  # pytype: disable=attribute-error  # class-cleanup
        if isinstance(value, collections.abc.Mapping):
            output_dict[key] = override(config_dict.get(key, {}), value)
        else:
            output_dict[key] = value
    return output_dict


class AxialResNetInstance(axial_resnet.AxialResNet):  # pytype: disable=ignored-abstractmethod  # abcmeta-check
    """A base Axial-ResNet model."""

    @classmethod
    @abc.abstractmethod
    def _get_config(cls):
        pass

    def __init__(self, name, **kwargs):
        """Builds an Axial-ResNet model."""
        # Get the config of the current model.
        current_config = self._get_config()

        # Override the default config with the current config. This line can be
        # omitted because the default config equals the default arguments of the
        # functions that build the model. But we make all the configs explicit here.
        current_config = override(_get_default_config(), current_config)

        # Finally, override the current model config with keyword arguments. In this
        # way, we still respect arguments passed as keyword arguments, such as
        # classification_mode, output_stride, etc.
        current_config = override(current_config, kwargs)
        logging.info('Axial-ResNet final config: %s', current_config)
        super(AxialResNetInstance, self).__init__(name, **current_config)


class MaXDeepLabS(AxialResNetInstance):
    """MaX-DeepLab-S for panoptic segmentation.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
      MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
        CVPR 2021. https://arxiv.org/abs/2012.00759
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        # Return an empty dictionary as the default values are all set for
        # MaX-DeepLab-S.
        return {}


class MaXDeepLabL(AxialResNetInstance):
    """MaX-DeepLab-L for panoptic segmentation.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
      MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
        CVPR 2021. https://arxiv.org/abs/2012.00759
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        return {
            'num_blocks': [3, 6, 3, 3],
            'backbone_type': 'wider_resnet',
            'backbone_use_transformer_beyond_stride': 16,
            'extra_decoder_use_transformer_beyond_stride': 16,
            'backbone_decoder_num_stacks': 1,
            'extra_decoder_num_stacks': 1,
            'extra_decoder_blocks_per_stage': 3,
            'memory_channels': 512,
            'base_transformer_expansion': 2.0,
            'global_feed_forward_network_channels': 512,
            'block_group_config': {
                'attention_bottleneck_expansion': 4,
                'drop_path_beyond_stride': 4,
                'axial_layer_config': {
                    'key_expansion': 2,
                    'value_expansion': 4,
                },
            },
        }


class MaXDeepLabSBackbone(MaXDeepLabS):
    """MaX-DeepLab-S backbone for image classification pretraining.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
      MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
        CVPR 2021. https://arxiv.org/abs/2012.00759
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(MaXDeepLabSBackbone, cls)._get_config()
        # Override the config of MaXDeepLabS.
        override_config = {
            'classification_mode': True,
            # The transformer blocks are not ImageNet pretrained. They are randomly
            # initialized and trained from scratch for panoptic segmentation.
            'backbone_use_transformer_beyond_stride': 0,
        }
        return override(base_config, override_config)


class MaXDeepLabLBackbone(MaXDeepLabL):
    """MaX-DeepLab-L backbone for image classification pretraining.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
      MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
        CVPR 2021. https://arxiv.org/abs/2012.00759
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(MaXDeepLabLBackbone, cls)._get_config()
        # Override the config of MaXDeepLabL.
        override_config = {
            'classification_mode': True,
            # The transformer blocks are not ImageNet pretrained. They are randomly
            # initialized and trained from scratch for panoptic segmentation.
            'backbone_use_transformer_beyond_stride': 0,
        }
        return override(base_config, override_config)


class ResNet50(AxialResNetInstance):
    """A ResNet-50 instance.

    Note that the implementation is different from the original ResNet-50 in:
      (1) We apply strided convolutions in the first 3x3 convolution of the first
          residual block of a stage.
      (2) We replace the strided max pooling layer in the stem by applying strided
          convolution in the immediate next residual block.
    """

    @classmethod
    def _get_config(cls):
        return {
            'classification_mode': True,
            'backbone_type': 'resnet',
            'use_axial_beyond_stride': 0,
            'backbone_use_transformer_beyond_stride': 0,
            'block_group_config': {
                'drop_path_keep_prob': 1.0,
            },
        }


class ResNet50Beta(ResNet50):
    """A ResNet-50 but with inception stem.

    Note that the implementation is different from the original ResNet-50 in:
      (1) We apply strided convolutions in the first 3x3 convolution of the first
          residual block of a stage.
      (2) We replace the strided max pooling layer in the stem by applying strided
          convolution in the immediate next residual block.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(ResNet50Beta, cls)._get_config()
        # Override the config of ResNet50.
        override_config = {
            'backbone_type': 'resnet_beta',
        }
        return override(base_config, override_config)


class AxialResNetL(ResNet50):
    """Axial-ResNet-L for image classification only.

    Axial-ResNet-L is a ResNet50 with use_axial_beyond_stride = 2.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(AxialResNetL, cls)._get_config()
        # Override the config of ResNet50.
        override_config = {
            'use_axial_beyond_stride': 2,
        }
        return override(base_config, override_config)


class AxialResNetS(ResNet50):
    """Axial-ResNet-S for image classification only.

    Axial-ResNet-S is a ResNet50 with use_axial_beyond_stride = 2 and
    width_multiplier = 0.5.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(AxialResNetS, cls)._get_config()
        # Override the config of ResNet50.
        override_config = {
            'width_multiplier': 0.5,
            'use_axial_beyond_stride': 2,
        }
        return override(base_config, override_config)


class AxialDeepLabL(ResNet50Beta):
    """Axial-DeepLab-L for panoptic segmentation.

    Axial-DeepLab-L is a ResNet50Beta with use_axial_beyond_stride = 2.
    Axial-DeepLab-L is also equivalent to Axial-ResNet-L with an inception stem.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(AxialDeepLabL, cls)._get_config()
        override_config = {
            'use_axial_beyond_stride': 2,
        }
        return override(base_config, override_config)


class AxialDeepLabS(ResNet50Beta):
    """Axial-DeepLab-S for panoptic segmentation.

    Axial-DeepLab-S is a ResNet50Beta with use_axial_beyond_stride = 2 and
    width_multiplier = 0.5.
    Axial-DeepLab-S is also equivalent to Axial-ResNet-S with an inception stem.

    Reference:
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(AxialDeepLabS, cls)._get_config()
        override_config = {
            'width_multiplier': 0.5,
            'use_axial_beyond_stride': 2,
        }
        return override(base_config, override_config)


class SWideRNet(AxialResNetInstance):
    """A SWideRNet instance.

    Note that the implementation is different from the original SWideRNet in:
      (1) We apply strided convolutions in the first residual block of a stage,
          instead of the last residual block.
      (2) We replace the strided max pooling layer in the stem by applying strided
          convolution in the immediate next residual block.
      (3) We (optionally) use squeeze and excitation in all five stages, instead
          of the last four stages only.

    Reference:
      Scaling Wide Residual Networks for Panoptic Segmentation,
        https://arxiv.org/abs/2011.11675
          Liang-Chieh Chen, Huiyu Wang, Siyuan Qiao.
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        return {
            'num_blocks': [3, 6, 3, 3],
            'classification_mode': True,
            'backbone_type': 'wider_resnet',
            'use_axial_beyond_stride': 0,
            'backbone_use_transformer_beyond_stride': 0,
            'block_group_config': {
                'drop_path_beyond_stride': 4,
                'conv_use_recompute_grad': True,
            },
        }


class AxialSWideRNet(SWideRNet):
    """SWideRNet with axial attention blocks in the last two stages.

    Note that the implementation is different from the original SWideRNet in:
      (1) We apply strided convolutions in the first residual block of a stage,
          instead of the last residual block.
      (2) We replace the strided max pooling layer in the stem by applying strided
          convolution in the immediate next residual block.
      (3) We (optionally) use squeeze and excitation in all five stages, instead
          of the last four stages only.

    Reference:
      Scaling Wide Residual Networks for Panoptic Segmentation,
        https://arxiv.org/abs/2011.11675
          Liang-Chieh Chen, Huiyu Wang, Siyuan Qiao.
      Axial-Deeplab: Stand-Alone Axial-Attention for Panoptic Segmentation,
        ECCV 2020 Spotlight. https://arxiv.org/abs/2003.07853
          Huiyu Wang, Yukun Zhu, Bradley Green, Hartwig Adam, Alan Yuille,
            Liang-Chieh Chen.
    """

    @classmethod
    def _get_config(cls):
        base_config = super(AxialSWideRNet, cls)._get_config()
        override_config = {
            'use_axial_beyond_stride': 16,
            'block_group_config': {
                'attention_bottleneck_expansion': 4,
                'axial_layer_config': {
                    'key_expansion': 2,
                    'value_expansion': 4,
                },
            },
        }
        return override(base_config, override_config)


def get_model(name, **kwargs):
    """Gets the model instance given the model name."""
    name_lower = name.lower()
    if name_lower == 'max_deeplab_s':
        return MaXDeepLabS(name_lower, **kwargs)
    elif name_lower == 'max_deeplab_l':
        return MaXDeepLabL(name_lower, **kwargs)
    elif name_lower == 'max_deeplab_s_backbone':
        return MaXDeepLabSBackbone(name_lower, **kwargs)
    elif name_lower == 'max_deeplab_l_backbone':
        return MaXDeepLabLBackbone(name_lower, **kwargs)
    elif name_lower == 'resnet50':
        return ResNet50(name_lower, **kwargs)
    elif name_lower == 'resnet50_beta':
        return ResNet50Beta(name_lower, **kwargs)
    elif name_lower == 'swidernet' or name_lower == 'wide_resnet41':
        return SWideRNet(name_lower, **kwargs)
    elif name_lower == 'axial_swidernet':
        return AxialSWideRNet(name_lower, **kwargs)
    elif name_lower == 'axial_resnet_s':
        return AxialResNetS(name_lower, **kwargs)
    elif name_lower == 'axial_resnet_l':
        return AxialResNetL(name_lower, **kwargs)
    elif name_lower == 'axial_deeplab_s':
        return AxialDeepLabS(name_lower, **kwargs)
    elif name_lower == 'axial_deeplab_l':
        return AxialDeepLabL(name_lower, **kwargs)
    else:
        raise ValueError(name_lower + ' is not supported.')
